

import sys
import math
#import bisect
from collections import defaultdict

import numpy
from scipy import integrate, interpolate
from ..integrators.statutils import cumvar, welford, update, finalize
import itertools
import functools


import os
if not( 'RIFT_LOWLATENCY'  in os.environ):
    # Dont support selected external packages in low latency
 try:
    import healpy
 except:
    print(" - No healpy - ")
 try:
    import vegas
 except:
    print(" - No vegas - ")


from multiprocessing import Pool


__author__ = "Chris Pankow <pankow@gravity.phys.uwm.edu>"

rosDebugMessages = True

class NanOrInf(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class MCSampler(object):
    """
    Class to define a set of parameter names, limits, and probability densities.
    """

    @staticmethod
    def match_params_from_args(args, params):
        """
        Given two unordered sets of parameters, one a set of all "basic" elements (strings) possible, and one a set of elements both "basic" strings and "combined" (basic strings in tuples), determine whether the sets are equivalent if no basic element is repeated.

        e.g. set A ?= set B

        ("a", "b", "c") ?= ("a", "b", "c") ==> True
        (("a", "b", "c")) ?= ("a", "b", "c") ==> True
        (("a", "b"), "d")) ?= ("a", "b", "c") ==> False  # basic element 'd' not in set B
        (("a", "b"), "d")) ?= ("a", "b", "d", "c") ==> False  # not all elements in set B represented in set A
        """
        not_common = set(args) ^ set(params)
        if len(not_common) == 0:
            # All params match
            return True
        if all([not isinstance(i, tuple) for i in not_common]):
            # The only way this is possible is if there are
            # no extraneous params in args
            return False

        to_match, against = [i for i in not_common if not isinstance(i, tuple)], [i for i in not_common if isinstance(i, tuple)]

        matched = []
        import itertools
        for i in range(2, max(list(map(len, against)))+1):
            matched.extend([t for t in itertools.permutations(to_match, i) if t in against])
        return (set(matched) ^ set(against)) == set()


    def __init__(self):
        # Total number of samples drawn
        self.ntotal = 0
        # Parameter names
        self.params = set()
        self.params_ordered = []  # keep them in order. Important to break likelihood function need for names
        # parameter -> pdf function object
        self.pdf = {}
        # If the pdfs aren't normalized, this will hold the normalization 
        # constant
        self._pdf_norm = defaultdict(lambda: 1)
        # Cache for the sampling points
        self._rvs = {}
        # parameter -> cdf^{-1} function object
        self.cdf = {}
        self.cdf_inv = {}
        # params for left and right limits
        self.llim, self.rlim = {}, {}
        # Keep track of the adaptive parameters
        self.adaptive = []

        # Keep track of the adaptive parameter 1-D marginalizations
        self._hist = {}

        # MEASURES (=priors): ROS needs these at the sampler level, to clearly separate their effects
        # ASSUMES the user insures they are normalized
        self.prior_pdf = {}

    def clear(self):
        """
        Clear out the parameters and their settings, as well as clear the sample cache.
        """
        self.params = set()
        self.params_ordered = []
        self.pdf = {}
        self._pdf_norm = defaultdict(lambda: 1.0)
        self._rvs = {}
        self._hist = {}
        self.cdf = {}
        self.cdf_inv = {}
        self.llim = {}
        self.rlim = {}
        self.adaptive = []

    def add_parameter(self, params, pdf,  cdf_inv=None, left_limit=None, right_limit=None, prior_pdf=None, adaptive_sampling=False):
        """
        Add one (or more) parameters to sample dimensions. params is either a string describing the parameter, or a tuple of strings. The tuple will indicate to the sampler that these parameters must be sampled together. left_limit and right_limit are on the infinite interval by default, but can and probably should be specified. If several params are given, left_limit, and right_limit must be a set of tuples with corresponding length. Sampling PDF is required, and if not provided, the cdf inverse function will be determined numerically from the sampling PDF.
        """
        self.params.add(params) # does NOT preserve order in which parameters are provided
        self.params_ordered.append(params)
        if rosDebugMessages: 
            print(" Adding parameter ", params, " with limits ", [left_limit, right_limit])
        if isinstance(params, tuple):
            assert all([lim[0] < lim[1] for lim in zip(left_limit, right_limit)])
            if left_limit is None:
                self.llim[params] = list(float("-inf"))*len(params)
            else:
                self.llim[params] = left_limit
            if right_limit is None:
                self.rlim[params] = list(float("+inf"))*len(params)
            else:
                self.rlim[params] = right_limit
        else:
            assert left_limit < right_limit
            if left_limit is None:
                self.llim[params] = float("-inf")
            else:
                self.llim[params] = left_limit
            if right_limit is None:
                self.rlim[params] = float("+inf")
            else:
                self.rlim[params] = right_limit
        self.pdf[params] = pdf
        # FIXME: This only works automagically for the 1d case currently
        self.cdf_inv[params] = cdf_inv or self.cdf_inverse(params)
        if not isinstance(params, tuple):
            self.cdf[params] =  self.cdf_function(params)
            if prior_pdf is None:
                self.prior_pdf[params] = lambda x:1
            else:
                self.prior_pdf[params] = prior_pdf
        else:
            self.prior_pdf[params] = prior_pdf

        if adaptive_sampling:
            print("   Adapting ", params)
            self.adaptive.append(params)


    def cdf_function(self, param):
        """
        Numerically determine the  CDF from a given sampling PDF. If the PDF itself is not normalized, the class will keep an internal record of the normalization and adjust the PDF values as necessary. Returns a function object which is the interpolated CDF.
        """
        # Solve P'(x) == p(x), with P[lower_boun] == 0
        def dP_cdf(p, x):
            if x > self.rlim[param] or x < self.llim[param]:
                return 0
            return numpy.float64(self.pdf[param](x))
        x_i = numpy.linspace(self.llim[param], self.rlim[param], 1000,dtype=numpy.float64)
        # Integrator needs to have a step size which doesn't step over the
        # probability mass
        # TODO: Determine h_max.
        cdf = integrate.odeint(dP_cdf, [0], x_i, hmax=0.01*numpy.float64(self.rlim[param]-self.llim[param])).T[0]  # issue of type limits for odeint
        if cdf[-1] != 1.0: # original pdf wasn't normalized
            self._pdf_norm[param] = cdf[-1]
            cdf /= cdf[-1]
        # Interpolate the inverse
        return interpolate.interp1d( x_i,cdf)

    def cdf_inverse(self, param):
        """
        Numerically determine the inverse CDF from a given sampling PDF. If the PDF itself is not normalized, the class will keep an internal record of the normalization and adjust the PDF values as necessary. Returns a function object which is the interpolated CDF inverse.
        """
        # Solve P'(x) == p(x), with P[lower_boun] == 0
        def dP_cdf(p, x):
            if x > self.rlim[param] or x < self.llim[param]:
                return 0
            return numpy.float64(self.pdf[param](x))
        x_i = numpy.linspace(self.llim[param], self.rlim[param], 1000,dtype=numpy.float64)
        # Integrator needs to have a step size which doesn't step over the
        # probability mass
        # TODO: Determine h_max.
        cdf = integrate.odeint(dP_cdf, [0], x_i, hmax=0.01*numpy.float64(self.rlim[param]-self.llim[param])).T[0]
        if cdf[-1] != 1.0: # original pdf wasn't normalized
            self._pdf_norm[param] = cdf[-1]
            cdf /= cdf[-1]
        # Interpolate the inverse
        return interpolate.interp1d(cdf, x_i)

    def draw(self, rvs, *args, **kwargs):
        """
        Draw a set of random variates for parameter(s) args. Left and right limits are handed to the function. If args is None, then draw *all* parameters. 'rdict' parameter is a boolean. If true, returns a dict matched to param name rather than list. rvs must be either a list of uniform random variates to transform for sampling, or an integer number of samples to draw.
        """
        if len(args) == 0:
            args = self.params

        no_cache_samples = kwargs["no_cache_samples"] if "no_cache_samples" in kwargs else False


        if isinstance(rvs, int) or isinstance(rvs, float):
            #
            # Convert all arguments to tuples
            #
            # FIXME: UGH! Really? This was the most elegant thing you could come
            # up with?
            rvs_tmp = [numpy.random.uniform(0,1,(len(p), int(rvs))) for p in [(i,) if not isinstance(i, tuple) else i for i in args]]
            rvs_tmp = numpy.array([self.cdf_inv[param](*rv) for (rv, param) in zip(rvs_tmp, args)], dtype=object)
        else:
            rvs_tmp = numpy.array(rvs)

        # FIXME: ELegance; get some of that...
        # This is mainly to ensure that the array can be "splatted", e.g.
        # separated out into its components for matching with args. The case of
        # one argument has to be handled specially.
        res = []
        for (cdf_rv, param) in zip(rvs_tmp, args):
            if len(cdf_rv.shape) == 1:
                res.append((self.pdf[param](numpy.float64(cdf_rv)).astype(numpy.float64)/self._pdf_norm[param], self.prior_pdf[param](cdf_rv), cdf_rv))
            else:
                # NOTE: the "astype" is employed here because the arrays can be
                # irregular and thus assigned the 'object' type. Since object
                # arrays can't be splatted, we have to force the conversion
                res.append((self.pdf[param](*cdf_rv.astype(numpy.float64))/self._pdf_norm[param], self.prior_pdf[param](*cdf_rv.astype(numpy.float64)), cdf_rv))

        #
        # Cache the samples we chose
        #
        if not no_cache_samples:  # more efficient memory usage. Note adaptation will not wor
          if len(self._rvs) == 0:
            self._rvs = dict(list(zip(args, rvs_tmp)))
          else:
            rvs_tmp = dict(list(zip(args, rvs_tmp)))
            #for p, ar in self._rvs.items():
            for p in self.params_ordered:
                self._rvs[p] = numpy.hstack( (self._rvs[p], rvs_tmp[p]) )
        else:  
            # if we are not caching samples, DELETE the sample record.  Saves memory!
            if len(self._rvs) >0:
                for p in self.params_ordered:
                    del self._rvs[p]
            self._rvs = {}

        #
        # Pack up the result if the user wants a dictonary instead
        #
        if "rdict" in kwargs:
            return dict(list(zip(args, res)))
        return list(zip(*res))


    def integrate_vegas(self, func, *args, **kwargs):
        """
        Uses vegas to do the integral.  Does not return sample points
        Remember:   pdf, cdf_inv refer to the *sampling* prior, so I need to multiply the integrand by a PDF ratio product!
        """
        # Method: use loop over params (=args), 
        #
        # Pin values
        #
        tempcdfdict, temppdfdict, temppriordict, temppdfnormdict = {}, {}, {}, {}
        temppdfnormdict = defaultdict(lambda: 1.0)
        for p, val in list(kwargs.items()):
            if p in self.params_ordered:
                # Store the previous pdf/cdf in case it's already defined
                tempcdfdict[p] = self.cdf_inv[p]
                temppdfdict[p] = self.pdf[p]
                temppdfnormdict[p] = self._pdf_norm[p]
                temppriordict[p] = self.prior_pdf[p]
                # Set a new one to always return the same value
                self.pdf[p] = functools.partial(delta_func_pdf_vector, val)
                self._pdf_norm[p] = 1.0
                self.prior_pdf[p] = functools.partial(delta_func_pdf_vector, val)
                self.cdf_inv[p] = functools.partial(delta_func_samp_vector, val)

        #
        # Determine stopping conditions
        #
        nmax = kwargs["nmax"] if "nmax" in kwargs else 1e6
        neff = kwargs["neff"] if "neff" in kwargs else 1000
        n = int(kwargs["n"]) if "n" in kwargs else min(1000, nmax)  # chunk size
        nBlocks = 10
        n_itr = numpy.max([10,numpy.min([20,int(nmax/nBlocks/n)])])  # largest number to use


        # What I am actually doing: an n-dimensional integral with vegas, using the CDF function to generate the parameter
        # range.
        paramListDefault = kwargs['param_order'] # do not try to get it from the function
        strToEval = "lambda x: func("
        for indx in numpy.arange(len(paramListDefault)):
            strToEval+= 'self.cdf_inv["'+str(paramListDefault[indx])+'"](x['+str(indx)+']),'
        strToEval=strToEval[:len(strToEval)-1]  # drop last comma
        strToEval += ')'
        # multiply by ratio of p/ps
        for indx in numpy.arange(len(paramListDefault)):
            strToEval+= '*( self.prior_pdf["'+str(paramListDefault[indx])+'"](x['+str(indx)+'])/(self.pdf["'+str(paramListDefault[indx])+'"](x['+str(indx)+'])/self._pdf_norm["'  +str(paramListDefault[indx])+ '"] ))'
        print(strToEval)
        fnToUse = eval(strToEval,{'func':func, 'self':self})  # evaluate in context
#        fnToUse =vegas.batchintegrand(fnToUse)   # batch mode
#        print fnToUse
#        grid = numpy.zeros((len(paramListDefault), 2))
#        grid[:,1] = numpy.ones(len(paramListDefault))
        integ = vegas.Integrator( len(paramListDefault)*[[0,1]]) # generate the grid
        # quick and dirty training
        print('Start training')
        result = integ(fnToUse,nitn=10, neval=1000)
        print(result.summary())
        # result -- problem of very highly peaked function, not clear if vanilla vegas is smart enough.
        # Loop over blocks of 1000 evaluations, and check that final chi2/dof  is within 0.05 of 1
        bDone = False
        alphaRunning = numpy.min([0.1, 8/numpy.log(result.mean)])   # constrain dynamic range to a reaonsable range
        print('Start full  : WARNING VEGAS TENDS TO OVERADAPT given huge dynamic range')
        while (not bDone and nBlocks):  # this is basically training
            print(" Block run " , n_itr, n)
            result=integ(fnToUse,nitn=n_itr, neval=n)
            alphaRunning = numpy.min([0.1, 8/numpy.log(result.mean)])   # constrain dynamic range to a reaonsable range
            print(nBlocks, numpy.sqrt(2*numpy.log(result.mean)), result.sdev/result.mean, result.chi2/result.dof, alphaRunning)
            print(result.summary())
            nBlocks+= -1
            if numpy.abs(result.chi2/result.dof - 1) < 0.05:
                bDone =True
        print(result.summary())
        return result

    #
    # FIXME: The priors are not strictly part of the MC integral, and so any
    # internal reference to them needs to be moved to a subclass which handles
    # the incovnenient part os doing the \int p/p_s L d\theta integral.
    #
    def integrate(self, func, *args, **kwargs):
        """
        Integrate func, by using n sample points. Right now, all params defined must be passed to args must be provided, but this will change soon.

        Limitations:
            func's signature must contain all parameters currently defined by the sampler, and with the same names. This is required so that the sample values can be passed consistently.

        kwargs:
        nmax -- total allowed number of sample points, will throw a warning if this number is reached before neff.
        neff -- Effective samples to collect before terminating. If not given, assume infinity
        n -- Number of samples to integrate in a 'chunk' -- default is 1000
        save_integrand -- Save the evaluated value of the integrand at the sample points with the sample point
        history_mult -- Number of chunks (of size n) to use in the adaptive histogramming: only useful if there are parameters with adaptation enabled
        tempering_exp -- Exponent to raise the weights of the 1-D marginalized histograms for adaptive sampling prior generation, by default it is 0 which will turn off adaptive sampling regardless of other settings
        temper_log -- Adapt in min(ln L, 10^(-5))^tempering_exp
        tempering_adapt -- Gradually evolve the tempering_exp based on previous history.
        floor_level -- *total probability* of a uniform distribution, averaged with the weighted sampled distribution, to generate a new sampled distribution
        n_adapt -- number of chunks over which to allow the pdf to adapt. Default is zero, which will turn off adaptive sampling regardless of other settings
        convergence_tests - dictionary of function pointers, each accepting self._rvs and self.params as arguments. CURRENTLY ONLY USED FOR REPORTING

        tripwire_fraction - fraction of nmax of iterations after which n_eff needs to be greater than 1+epsilon for epsilon a small number
        tripwire_epsilon - small number used in tripwire test

        Pinning a value: By specifying a kwarg with the same of an existing parameter, it is possible to "pin" it. The sample draws will always be that value, and the sampling prior will use a delta function at that value.
        """

        #
        # Pin values
        #
        n_horrible = 0
        n_horrible_max = 10

        tempcdfdict, temppdfdict, temppriordict, temppdfnormdict = {}, {}, {}, {}
        temppdfnormdict = defaultdict(lambda: 1.0)
        for p, val in list(kwargs.items()):
            if p in self.params_ordered:  
                # Store the previous pdf/cdf in case it's already defined
                tempcdfdict[p] = self.cdf_inv[p]
                temppdfdict[p] = self.pdf[p]
                temppdfnormdict[p] = self._pdf_norm[p]
                temppriordict[p] = self.prior_pdf[p]
                # Set a new one to always return the same value
                self.pdf[p] = functools.partial(delta_func_pdf_vector, val)
                self._pdf_norm[p] = 1.0
                self.prior_pdf[p] = functools.partial(delta_func_pdf_vector, val)
                self.cdf_inv[p] = functools.partial(delta_func_samp_vector, val)

        # put it back in the args
        #args = tuple(list(args) + filter(lambda p: p in self.params, kwargs.keys()))
        # This is a semi-hack to ensure that the integrand is called with
        # the arguments in the right order
        # FIXME: How dangerous is this?
        no_protect_names = False
        if 'no_protect_names' in kwargs:
           no_protect_names = kwargs['no_protect_names']
        if hasattr(func,'__code__') and not(no_protect_names):
            args = func.__code__.co_varnames[:func.__code__.co_argcount]
        else:
            args=list(args[:len(args)])
            no_protect_params=True   # not code dictionry, so just use arguments in order

        #if set(args) & set(params) != set(args):
        # DISABLE THIS CHECK
#        if not MCSampler.match_params_from_args(args, self.params):
#            raise ValueError("All integrand variables must be represented by integral parameters.")
        
        #
        # Determine stopping conditions
        #
        nmax = int(kwargs["nmax"]) if "nmax" in kwargs else float("inf")
        neff = kwargs["neff"] if "neff" in kwargs else numpy.float128("inf")
        n = int(kwargs["n"]) if "n" in kwargs else min(1000, nmax)
        convergence_tests = kwargs["convergence_tests"] if "convergence_tests" in kwargs else None


        #
        # Adaptive sampling parameters
        #
        n_history = int(kwargs["history_mult"]*n) if "history_mult" in kwargs else n
        tempering_exp = kwargs["tempering_exp"] if "tempering_exp" in kwargs else 0.0
        n_adapt = int(kwargs["n_adapt"]*n) if "n_adapt" in kwargs else 0
        floor_integrated_probability = kwargs["floor_level"] if "floor_level" in kwargs else 0
        temper_log = kwargs["tempering_log"] if "tempering_log" in kwargs else False
        tempering_adapt = kwargs["tempering_adapt"] if "tempering_adapt" in kwargs else False
        if not tempering_adapt:
            tempering_exp_running=tempering_exp
        else:
            print(" Adaptive tempering ")
            #tempering_exp_running=0.01  # decent place to start for the first step. Note starting at zero KEEPS it at zero.
            tempering_exp_running=tempering_exp
            

        save_intg = kwargs["save_intg"] if "save_intg" in kwargs else False
        force_no_adapt = kwargs["force_no_adapt"] if "force_no_adapt" in kwargs else False
        save_no_samples = kwargs["save_no_samples"] if "save_no_samples" in kwargs else False
        if save_no_samples:   # can't adapt without saved samples
            force_no_adapt = True
        # FIXME: The adaptive step relies on the _rvs cache, so this has to be
        # on in order to work
        if n_adapt > 0 and tempering_exp > 0.0:
            save_intg = True

        deltalnL = kwargs['igrand_threshold_deltalnL'] if 'igrand_threshold_deltalnL' in kwargs else float("Inf") # default is to return all
        deltaP    = kwargs["igrand_threshold_p"] if 'igrand_threshold_p' in kwargs else 0 # default is to omit 1e-7 of probability
        bFairdraw  = kwargs["igrand_fairdraw_samples"] if "igrand_fairdraw_samples" in kwargs else False
        n_extr = kwargs["igrand_fairdraw_samples_max"] if "igrand_fairdraw_samples_max" in kwargs else None

        tripwire_fraction = kwargs["tripwire_fraction"] if "tripwire_fraction" in kwargs else 2  # make it impossible to trigger
        tripwire_epsilon = kwargs["tripwire_epsilon"] if "tripwire_epsilon" in kwargs else 0.001 # if we are not reasonably far away from unity, fail!

        bUseMultiprocessing = kwargs['use_multiprocessing'] if 'use_multiprocessing' in kwargs else False
        nProcesses = kwargs['nprocesses'] if 'nprocesses' in kwargs else 2
        bShowEvaluationLog = kwargs['verbose'] if 'verbose' in kwargs else False
        bShowEveryEvaluation = kwargs['extremely_verbose'] if 'extremely_verbose' in kwargs else False

        if bShowEvaluationLog:
            print(" .... mcsampler : providing verbose output ..... ")
        if bUseMultiprocessing:
            if rosDebugMessages:
                print(" Initiating multiprocessor pool : ", nProcesses)
            p = Pool(nProcesses)

        int_val1 = numpy.float128(0)
        self.ntotal = 0
        maxval = -float("Inf")
        maxlnL = -float("Inf")
        eff_samp = 0
        mean, var = None, numpy.float128(0)    # to prevent infinite variance due to overflow

        if bShowEvaluationLog:
            print("iteration Neff  sqrt(2*lnLmax) sqrt(2*lnLmarg) ln(Z/Lmax) int_var")

        if convergence_tests:
            bConvergenceTests = False   # start out not converged, if tests are on
            last_convergence_test = defaultdict(lambda: False)   # initialize record of tests
        else:
            bConvergenceTests = False    # if tests are not available, assume not converged. The other criteria will stop it
            last_convergence_test = defaultdict(lambda: False)   # need record of tests to be returned always
        while (eff_samp < neff and self.ntotal < nmax): #  and (not bConvergenceTests):
            if (self.ntotal > nmax*tripwire_fraction) and (eff_samp < 1+tripwire_epsilon):
                print(" Tripwire: n_eff too low ")
                raise Exception("Tripwire on n_eff")

            if n_horrible >= n_horrible_max:
                raise Exception("mcsampler: Too many iterations with no "
                                "contribution to integral, hard fail")

            # Draw our sample points
            args_draw ={}
            if force_no_adapt or save_no_samples:  # don't save permanent sample history if not needed
                args_draw.update({"no_cache_samples":True})
            p_s, p_prior, rv = self.draw(n, *self.params_ordered,**args_draw)  # keep in order

#            print "Prior ",  type(p_prior[0]), p_prior[0]
#            print "rv ",  type(rv[0]), rv[0]    # rv generally has dtype = object, to enable joint sampling with multid variables
#            print "Sampling prior ", type(p_s[0]), p_s[0]  
                        
            # Calculate the overall p_s assuming each pdf is independent
            joint_p_s = numpy.prod(p_s, axis=0)
            joint_p_prior = numpy.prod(p_prior, axis=0)
            joint_p_prior = numpy.array(joint_p_prior,dtype=numpy.float128)  # Force type. Some type issues have arisen (dtype=object returns by accident)

#            print "Joint prior ",  type(joint_p_prior), joint_p_prior.dtype, joint_p_prior
#            print "Joint sampling prior ", type(joint_p_s), joint_p_s.dtype

            #
            # Prevent zeroes in the sampling prior
            #
            # FIXME: If we get too many of these, we should bail
            if any(joint_p_s <= 0):
                for p in self.params_ordered:
                    self._rvs[p] = numpy.resize(self._rvs[p], len(self._rvs[p])-n)
                print("Zero prior value detected, skipping.", file=sys.stderr)
                n_horrible += 1
                continue

            #
            # Unpack rvs and evaluate integrand
            #
            if len(rv[0].shape) != 1:
                rv = rv[0]

            params = []
            for item in self.params_ordered:  # USE IN ORDER
                if isinstance(item, tuple):
                    params.extend(item)
                else:
                    params.append(item)
            unpacked = unpacked0 = numpy.hstack([r.flatten() for r in rv]).reshape(len(args), -1)
            unpacked = dict(list(zip(params, unpacked)))
            if 'no_protect_names' in kwargs:
                fval = func(*unpacked0)  # do not protect order
            else:
                fval = func(**unpacked) # Chris' original plan: note this insures the function arguments are tied to the parameters, using a dictionary. 

            #
            # Check if there is any practical contribution to the integral
            #
            # FIXME: While not technically a fatal error, this will kill the
            # adaptive sampling
            if fval.sum() == 0:
                for p in self.params_ordered:
                    self._rvs[p] = numpy.resize(self._rvs[p], len(self._rvs[p])-n)
                print("No contribution to integral, skipping.", file=sys.stderr)
                n_horrible += 1
                continue

            if save_intg and not force_no_adapt:
                # FIXME: The joint_prior, if not specified is set to one and
                # will come out as a scalar here, hence the hack
                if not isinstance(joint_p_prior, numpy.ndarray):
                    joint_p_prior = numpy.ones(fval.shape)*joint_p_prior


#                print "      Prior", type(joint_p_prior), joint_p_prior.dtype

                # FIXME: See warning at beginning of function. The prior values
                # need to be moved out of this, as they are not part of MC
                # integration
                if "integrand" in self._rvs:
                    self._rvs["integrand"] = numpy.hstack( (self._rvs["integrand"], fval) )
                    self._rvs["joint_prior"] = numpy.hstack( (self._rvs["joint_prior"], joint_p_prior) )
                    self._rvs["joint_s_prior"] = numpy.hstack( (self._rvs["joint_s_prior"], joint_p_s) )
                    self._rvs["weights"] = numpy.hstack( (self._rvs["joint_s_prior"], fval*joint_p_prior/joint_p_s) )
                else:
                    self._rvs["integrand"] = fval
                    self._rvs["joint_prior"] = joint_p_prior
                    self._rvs["joint_s_prior"] = joint_p_s
                    self._rvs["weights"] = fval*joint_p_prior/joint_p_s

            # Calculate the integral over this chunk
            int_val = fval * joint_p_prior / joint_p_s

            if bShowEveryEvaluation:
                for i in range(n):
                    print(" Evaluation details: p,ps, L = ", joint_p_prior[i], joint_p_s[i], fval[i])
                    print(self.params_ordered)
                    for indx in numpy.arange(len(args)):
                        print( self.params_ordered[indx],p_s[indx],p_prior[indx],rv[indx])

            # Calculate max L (a useful convergence feature) for debug 
            # reporting.  Not used for integration
            # Try to avoid nan's
            maxlnL = numpy.log(numpy.max([numpy.exp(maxlnL), numpy.max(fval),numpy.exp(-100)]))   # note if f<0, this will return nearly 0

            # Calculate the effective samples via max over the current 
            # evaluations
            maxval = [max(maxval, int_val[0]) if int_val[0] != 0 else maxval]
            for v in int_val[1:]:
                maxval.append( v if v > maxval[-1] and v != 0 else maxval[-1] )

            # running variance
#            var = cumvar(int_val, mean, var, int(self.ntotal))[-1]
#            var = welford(int_val, mean, var, int(self.ntotal))
            if var is None:
                var=0
            if mean is None:
                mean=0
            current_aggregate = [int(self.ntotal),mean, (self.ntotal-1)*var]
            current_aggregate = update(current_aggregate, int_val)
            outvals = finalize(current_aggregate)
#            print(var, outvals[-1])
            var = outvals[-1]
            # running integral (note also in current_aggregate)
            int_val1 += int_val.sum()
            # running number of evaluations
            self.ntotal += n
            # FIXME: Likely redundant with int_val1
            mean = int_val1/self.ntotal
            maxval = maxval[-1]

            eff_samp = int_val1/maxval

            # Throw exception if we get infinity or nan
            if math.isnan(eff_samp):
                raise NanOrInf("Effective samples = nan")
            if maxlnL is float("Inf"):
                raise NanOrInf("maxlnL = inf")

            if bShowEvaluationLog:
                print(" :",  self.ntotal, eff_samp, numpy.sqrt(2*maxlnL), numpy.sqrt(2*numpy.log(int_val1/self.ntotal)), numpy.log(int_val1/self.ntotal)-maxlnL, numpy.sqrt(var*self.ntotal)/int_val1)

            if (not convergence_tests) and self.ntotal >= nmax and neff != float("inf"):
                print("WARNING: User requested maximum number of samples reached... bailing.", file=sys.stderr)

            # Convergence tests:
            if convergence_tests:
                bConvergedThisIteration = True  # start out optimistic
                for key in list(convergence_tests.keys()):
                    last_convergence_test[key] =  convergence_tests[key](self._rvs, self.params_ordered)
                    bConvergedThisIteration = bConvergedThisIteration  and                      last_convergence_test[key]
                bConvergenceTests = bConvergedThisIteration

            if convergence_tests  and bShowEvaluationLog:  # Print status of each test
                for key in convergence_tests:
                    print("   -- Convergence test status : ", key, last_convergence_test[key])

            #
            # The total number of adaptive steps is reached
            #
            # FIXME: We need a better stopping condition here
            if self.ntotal > n_adapt:
               if bShowEvaluationLog:
                  print(" ... skipping adaptation in late iterations .. ")
               continue

            if force_no_adapt:
               continue

            # FIXME: Hardcoding
            #mixing_floor = 10**(-numpy.sqrt(ntotal))
            #mixing_floor = 10**-50

            #
            # Iterate through each of the parameters, updating the sampling
            # prior PDF according to the 1-D marginalization
            #
            for itr, p in enumerate(self.params_ordered):
                # FIXME: The second part of this condition should be made more
                # specific to pinned parameters
                if p not in self.adaptive or p in list(kwargs.keys()):
                    continue
                points = self._rvs[p][-n_history:]
#                print "      Points", p, type(points),points.dtype
                # use log weights or weights
                if not temper_log:
                    weights = (self._rvs["integrand"][-n_history:]/self._rvs["joint_s_prior"][-n_history:]*self._rvs["joint_prior"][-n_history:])**tempering_exp_running
                else:
                    weights = numpy.maximum(1e-5,numpy.log(self._rvs["integrand"][-n_history:] )) #**tempering_exp_running

                if tempering_adapt:
                    # have the adaptive exponent converge to 2/ln(w_max), ln (w)*alpha <= 2. This helps dynamic range
                    # almost always dominated by the parameters we care about
                    tempering_exp_running = 0.8 *tempering_exp_running + 0.2*(3./numpy.max([1,numpy.log(numpy.max(weights))]))
                    if rosDebugMessages:
                        print("     -  New adaptive exponent  ", tempering_exp_running, " based on max 1d weight ", numpy.max(weights), " based on parameter ", p)

#                print "      Weights",  type(weights),weights.dtype
                self._hist[p], edges = numpy.histogram( points,
                    bins = 100,
                    range = (self.llim[p], self.rlim[p]),
                    weights = weights
                )
                # FIXME: numpy.hist can't normalize worth a damn
                self._hist[p] /= self._hist[p].sum()

                # Mix with uniform distribution
                self._hist[p] = (1-floor_integrated_probability)*self._hist[p] + numpy.ones(len(self._hist[p]))*floor_integrated_probability/len(self._hist[p])
                if rosDebugMessages and bShowEvaluationLog:
                    print("         Weight entropy (after histogram) ", numpy.sum(-1*self._hist[p]*numpy.log(self._hist[p])), p)

                edges = [ (e0+e1)/2.0 for e0, e1 in zip(edges[:-1], edges[1:]) ]
                edges.append( edges[-1] + (edges[-1] - edges[-2]) )
                edges.insert( 0, edges[0] - (edges[-1] - edges[-2]) )

                # FIXME: KS test probably has a place here. Not sure yet where.
                #from scipy.stats import kstest
                #d, pval = kstest(self._rvs[p][-n_history:], self.cdf[p])
                #print p, d, pval

                self.pdf[p] = interpolate.interp1d(edges, [0] + list(self._hist[p]) + [0])
                self.cdf[p] = self.cdf_function(p)
                self.cdf_inv[p] = self.cdf_inverse(p)

        # If we were pinning any values, undo the changes we did before
        self.cdf_inv.update(tempcdfdict)
        self.pdf.update(temppdfdict)
        self._pdf_norm.update(temppdfnormdict)
        self.prior_pdf.update(temppriordict)

        # Clean out the _rvs arrays for 'irrelevant' points
        #   - find and remove samples with  lnL less than maxlnL - deltalnL (latter user-specified)
        #   - create the cumulative weights
        #   - find and remove samples which contribute too little to the cumulative weights
        if "integrand" in self._rvs:
            self._rvs["sample_n"] = numpy.arange(len(self._rvs["integrand"]))  # create 'iteration number'        
            # Step 1: Cut out any sample with lnL belw threshold
            indx_list = [k for k, value in enumerate( (self._rvs["integrand"] > maxlnL - deltalnL)) if value] # threshold number 1
            # FIXME: This is an unncessary initial copy, the second step (cum i
            # prob) can be accomplished with indexing first then only pare at
            # the end
            for key in list(self._rvs.keys()):
                if isinstance(key, tuple):
                    self._rvs[key] = self._rvs[key][:,indx_list]
                else:
                    self._rvs[key] = self._rvs[key][indx_list]
            # Step 2: Create and sort the cumulative weights, among the remaining points, then use that as a threshold
            wt = self._rvs["integrand"]*self._rvs["joint_prior"]/self._rvs["joint_s_prior"]
            idx_sorted_index = numpy.lexsort((numpy.arange(len(wt)), wt))  # Sort the array of weights, recovering index values
            indx_list = numpy.array( [[k, wt[k]] for k in idx_sorted_index])     # pair up with the weights again. NOTE NOT INTEGER TYPE ANY MORE
            cum_sum = numpy.cumsum(indx_list[:,1])  # find the cumulative sum
            cum_sum = cum_sum/cum_sum[-1]          # normalize the cumulative sum
            indx_list = [int(indx_list[k, 0]) for k, value in enumerate(cum_sum > deltaP) if value]  # find the indices that preserve > 1e-7 of total probability. RECAST TO INTEGER

            # FIXME: See previous FIXME
            for key in list(self._rvs.keys()):
                if isinstance(key, tuple):
                    self._rvs[key] = self._rvs[key][:,indx_list]
                else:
                    self._rvs[key] = self._rvs[key][indx_list]

        # Do a fair draw of points, if option is set
        if bFairdraw and not(n_extr is None):
           n_extr = int(numpy.min([n_extr,1.5*eff_samp,1.5*neff]))
           print(" Fairdraw size : ", n_extr)
           wt = numpy.array(self._rvs["integrand"]*self._rvs["joint_prior"]/self._rvs["joint_s_prior"]/numpy.max(self._rvs["integrand"]),dtype=float)
           wt *= 1.0/numpy.sum(wt)
           if n_extr < len(self._rvs["integrand"]):
               indx_list = numpy.random.choice(numpy.arange(len(wt)), size=n_extr,replace=True,p=wt) # fair draw
               # FIXME: See previous FIXME
               for key in list(self._rvs.keys()):
                   if isinstance(key, tuple):
                       self._rvs[key] = self._rvs[key][:,indx_list]
                   else:
                       self._rvs[key] = self._rvs[key][indx_list]

        # Create extra dictionary to return things
        dict_return ={}
        if convergence_tests is not None:
            dict_return["convergence_test_results"] = last_convergence_test

        return int_val1/self.ntotal, var/self.ntotal, eff_samp, dict_return

### UTILITIES: Predefined distributions
#  Be careful: vectorization is not always implemented consistently in new versions of numpy
def uniform_samp(a, b, x):   # I prefer to vectorize with the same call for all functions, rather than hardcode vectorization
        if  x>a and x<b:
                return 1.0/(b-a)
        else:
                return 0
def uniform_samp_cdf_inv_vector(a,b,p):
    # relies on input being a numpy array, with range from 0 to 1
    out= p.copy()
    out = p*(b-a) + a
    return out
#uniform_samp_vector = numpy.vectorize(uniform_samp,excluded=['a','b'],otypes=[numpy.float])
uniform_samp_vector = numpy.vectorize(uniform_samp,otypes=[numpy.float64])

def ret_uniform_samp_vector_alt(a,b):
    return lambda x: numpy.where( (x>a) & (x<b), numpy.reciprocal(b-a),0.0)

# def uniform_samp_withfloor_vector(rmaxQuad,rmaxFlat,pFlat,x):
#     ret =0.
#     if x<rmaxQuad:
#         ret+= (1-pFlat)/rmaxQuad
#     if x<rmaxFlat:
#         ret +=pFlat/rmaxFlat
#     return  ret
# uniform_samp_withfloor_vector = numpy.vectorize(uniform_samp_withfloor, otypes=[numpy.float])
def uniform_samp_withfloor_vector(rmaxQuad,rmaxFlat,pFlat,x):
    if isinstance(x, float):
        ret =0.
        if x<rmaxQuad:
            ret+= (1-pFlat)/rmaxQuad
        if x<rmaxFlat:
            ret +=pFlat/rmaxFlat
        return  ret
    ret = numpy.zeros(x.shape,dtype=numpy.float64)
    ret += numpy.select([x<rmaxQuad],[(1.-pFlat)/rmaxQuad])
    ret += numpy.select([x<rmaxFlat],[pFlat/rmaxFlat])
    return ret



# syntatic sugar : predefine the most common distributions
uniform_samp_phase = lambda x,numpy=numpy: numpy.broadcast_to(0.5/numpy.pi, numpy.shape(x))
uniform_samp_psi = lambda x,numpy=numpy: numpy.broadcast_to(1.0/numpy.pi, numpy.shape(x))
uniform_samp_theta = lambda x,numpy=numpy: 0.5*numpy.sin(x.astype(float))
uniform_samp_dec = lambda x,numpy=numpy: 0.5*numpy.cos(x.astype(float))

uniform_samp_cos_theta = lambda x: numpy.broadcast_to(0.5, numpy.shape(x))
# uniform_samp_phase = numpy.vectorize(lambda x: 1/(2*numpy.pi))
# uniform_samp_psi = numpy.vectorize(lambda x: 1/(numpy.pi))
# uniform_samp_theta = numpy.vectorize(lambda x: numpy.sin(x)/(2))
# uniform_samp_dec = numpy.vectorize(lambda x: numpy.cos(x)/(2))

# uniform_samp_cos_theta = numpy.vectorize(lambda x: 1./2.)  #dumbest-possible implementation

def quadratic_samp(rmax,x):
        if x<rmax:
                return x**2/(3*rmax**3)
        else:
                return 0

quadratic_samp_vector = numpy.vectorize(quadratic_samp, otypes=[numpy.float64])

def inv_uniform_cdf(a, b, x):
    return (b-a)*x+a

def gauss_samp(mu, std, x):
    return 1.0/numpy.sqrt(2*numpy.pi*std**2)*numpy.exp(-(x-mu)**2/2/std**2)

def gauss_samp_withfloor(mu, std, myfloor, x):
    return 1.0/numpy.sqrt(2*numpy.pi*std**2)*numpy.exp(-(x-mu)**2/2/std**2) + myfloor

#gauss_samp_withfloor_vector = numpy.vectorize(gauss_samp_withfloor,excluded=['mu','std','myfloor'],otypes=[numpy.float])
gauss_samp_withfloor_vector = numpy.vectorize(gauss_samp_withfloor,otypes=[numpy.float64])


# Mass ratio. PDF propto 1/(1+q)^2.  Defined so mass ratio is < 1
# expr = Integrate[1/(1 + q)^2, q]
# scale = (expr /. q -> qmax )  - (expr /. q -> qmin)
# (expr - (expr /. q -> qmin))/scale == x // Simplify
# q /. Solve[%, q][[1]] // Simplify
# % // CForm
def q_samp_vector(qmin,qmax,x):
    scale = 1./(1+qmin) - 1./(1+qmax)
    return 1/numpy.power((1+x),2)/scale
def q_cdf_inv_vector(qmin,qmax,x):
    return numpy.array((qmin + qmax*qmin + qmax*x - qmin*x)/(1 + qmax - qmax*x + qmin*x),dtype=np.float128)

# total mass. Assumed used with q.  2M/Mmax^2-Mmin^2
def M_samp_vector(Mmin,Mmax,x):
    scale = 2./(Mmax**2 - Mmin**2)
    return x*scale


def cos_samp(x):
        return numpy.sin(x)/2   # x from 0, pi

def dec_samp(x):
        return numpy.sin(x+numpy.pi/2)/2   # x from 0, pi


cos_samp_vector = lambda x: cos_samp(numpy.array(x,dtype=numpy.float64))
dec_samp_vector = lambda x: dec_samp(numpy.array(x,dtype=numpy.float64))

#cos_samp_vector = numpy.vectorize(cos_samp,otypes=[numpy.float])
#dec_samp_vector = numpy.vectorize(dec_samp,otypes=[numpy.float])
def cos_samp_cdf_inv_vector(p):
    return numpy.arccos( 2*p-1)   # returns from 0 to pi
def dec_samp_cdf_inv_vector(p):
    return numpy.arccos(2*p-1) - numpy.pi/2  # target from -pi/2 to pi/2


def pseudo_dist_samp(r0,r):
        return r*r*numpy.exp( - (r0/r)*(r0/r)/2. + r0/r)+0.01  # put a floor on probability, so we converge. Note this floor only cuts out NEARBY distances

#pseudo_dist_samp_vector = numpy.vectorize(pseudo_dist_samp,excluded=['r0'],otypes=[numpy.float])
pseudo_dist_samp_vector = numpy.vectorize(pseudo_dist_samp,otypes=[numpy.float64])

def delta_func_pdf(x_0, x):
    return 1.0 if x == x_0 else 0.0

delta_func_pdf_vector = numpy.vectorize(delta_func_pdf, otypes=[numpy.float64])

def delta_func_samp(x_0, x):
    return x_0

delta_func_samp_vector = numpy.vectorize(delta_func_samp, otypes=[numpy.float64])


def linear_down_samp(x,xmin=0,xmax=1):
    """
    distribution p(x) \propto (xmax-x) 
    """
    return 2./(xmax-xmin)**2 * numpy.power(xmax-x,1)

def linear_down_samp_cdf(x,xmin=0,xmax=1):
    r"""
    CDF of distribution p(x) \propto (xmax-x) 
    """
    return 1.-1./(xmax-xmin)**2 * numpy.power(xmax-x,2)


def power_down_samp(x,xmin=0,xmax=1,alpha=3):
    r"""
    distribution p(x) \propto (xmax-x) 
    """
    return alpha/(xmax-xmin)**alpha * numpy.power(xmax-x,alpha-1)

def power_down_samp_cdf(x,xmin=0,xmax=1,alpha=3):
    r"""
    CDF of distribution p(x) \propto (xmax-x) 
    """
    return 1.-1./(xmax-xmin)**alpha * numpy.power(xmax-x,alpha)


class HealPixSampler(object):
    """
    Class to sample the sky using a FITS healpix map. Equivalent to a joint 2-D pdf in RA and dec.
    """

    @staticmethod
    def thph2decra(th, ph):
        """
        theta/phi to RA/dec
        theta (north to south) (0, pi)
        phi (east to west) (0, 2*pi)
        declination: north pole = pi/2, south pole = -pi/2
        right ascension: (0, 2*pi)
        
        dec = pi/2 - theta
        ra = phi
        """
        return numpy.pi/2-th, ph

    @staticmethod
    def decra2thph(dec, ra):
        """
        theta/phi to RA/dec
        theta (north to south) (0, pi)
        phi (east to west) (0, 2*pi)
        declination: north pole = pi/2, south pole = -pi/2
        right ascension: (0, 2*pi)
        
        theta = pi/2 - dec
        ra = phi
        """
        return numpy.pi/2-dec, ra

    def __init__(self, skymap, massp=1.0):
        self.skymap = skymap
        self._massp = massp
        self.renormalize()

    @property
    def massp(self):
        return self._massp

    @massp.setter
    def massp(self, value):
        assert 0 <= value <= 1
        self._massp = value
        norm = self.renormalize()

    def renormalize(self):
        """
        Identify the points contributing to the overall cumulative probability distribution, and set the proper normalization.
        """
        res = healpy.npix2nside(len(self.skymap))
        self.pdf_sorted = sorted([(p, i) for i, p in enumerate(self.skymap)], reverse=True)
        self.valid_points_decra = []
        cdf, np = 0, 0
        for p, i in self.pdf_sorted:
            if p == 0:
                continue # Can't have a zero prior
            self.valid_points_decra.append(HealPixSampler.thph2decra(*healpy.pix2ang(res, i)))
            cdf += p
            if cdf > self._massp:
                break
        self._renorm = cdf
        # reset to indicate we'd need to recalculate this
        self.valid_points_hist = None
        return self._renorm

    def __expand_valid(self, min_p=1e-7):
        #
        # Determine what the 'quanta' of probabilty is
        #
        if self._massp == 1.0:
            # This is to ensure we don't blow away everything because the map
            # is very spread out
            min_p = min(min_p, max(self.skymap))
        else:
            # NOTE: Only valid if CDF descending order is kept
            min_p = self.pseudo_pdf(*self.valid_points_decra[-1])

        self.valid_points_hist = []
        ns = healpy.npix2nside(len(self.skymap))

        # Renormalize first so that the vector histogram is properly normalized
        self._renorm = 0
        # Account for probability lost due to cut off
        for i, v in enumerate(self.skymap >= min_p):
            self._renorm += self.skymap[i] if v else 0

        for pt in self.valid_points_decra:
            th, ph = HealPixSampler.decra2thph(pt[0], pt[1])
            pix = healpy.ang2pix(ns, th, ph)
            if self.skymap[pix] < min_p:
                continue
            self.valid_points_hist.extend([pt]*int(round(self.pseudo_pdf(*pt)/min_p)))
        self.valid_points_hist = numpy.array(self.valid_points_hist).T

    def pseudo_pdf(self, dec_in, ra_in):
        """
        Return pixel probability for a given dec_in and ra_in. Note, uses healpy functions to identify correct pixel.
        """
        th, ph = HealPixSampler.decra2thph(dec_in, ra_in)
        res = healpy.npix2nside(len(self.skymap))
        return self.skymap[healpy.ang2pix(res, th, ph)]/self._renorm

    def pseudo_cdf_inverse(self, dec_in=None, ra_in=None, ndraws=1, stype='vecthist'):
        """
        Select points from the skymap with a distribution following its corresponding pixel probability. If dec_in, ra_in are suupplied, they are ignored except that their shape is reproduced. If ndraws is supplied, that will set the shape. Will return a 2xN numpy array of the (dec, ra) values.
        stype controls the type of sampling done to retrieve points. Valid choices are
        'rejsamp': Rejection sampling: accurate but slow
        'vecthist': Expands a set of points into a larger vector with the multiplicity of the points in the vector corresponding roughly to the probability of drawing that point. Because this is not an exact representation of the proability, some points may not be represented at all (less than quantum of minimum probability) or inaccurately (a significant fraction of the fundamental quantum).
        """

        if ra_in is not None:
            ndraws = len(ra_in)
        if ra_in is None:
            ra_in, dec_in = numpy.zeros((2, ndraws))

        if stype == 'rejsamp':
            # FIXME: This is only valid under descending ordered CDF summation
            ceiling = max(self.skymap)
            i, np = 0, len(self.valid_points_decra)
            while i < len(ra_in):
                rnd_n = numpy.random.randint(0, np)
                trial = numpy.random.uniform(0, ceiling)
                if trial <= self.pseudo_pdf(*self.valid_points_decra[rnd_n]):
                    dec_in[i], ra_in[i] = self.valid_points_decra[rnd_n]
                    i += 1
            return numpy.array([dec_in, ra_in])
        elif stype == 'vecthist':
            if self.valid_points_hist is None:
                self.__expand_valid()
            np = self.valid_points_hist.shape[1]
            rnd_n = numpy.random.randint(0, np, len(ra_in))
            dec_in, ra_in = self.valid_points_hist[:,rnd_n]
            return numpy.array([dec_in, ra_in])
        else:
            raise ValueError("%s is not a recgonized sampling type" % stype)

#pseudo_dist_samp_vector = numpy.vectorize(pseudo_dist_samp,excluded=['r0'],otypes=[numpy.float])
pseudo_dist_samp_vector = numpy.vectorize(pseudo_dist_samp,otypes=[numpy.float64])


def sanityCheckSamplerIntegrateUnity(sampler,*args,**kwargs):
        return sampler.integrate(lambda *args: 1,*args,**kwargs)

###
### CONVERGENCE TESTS
###


# neff by another name:
#    - value: tests for 'smooth' 1-d cumulative distributions
#    - require  require the most significant-weighted point be less than p of all cumulative probability
#    - this test is *equivalent* to neff > 1/p
#    - provided to illustrate the interface
def convergence_test_MostSignificantPoint(pcut, rvs, params):
    weights = rvs["weights"] #rvs["integrand"]* rvs["joint_prior"]/rvs["joint_s_prior"]
    indxmax = numpy.argmax(weights)
    wtSum = numpy.sum(weights)
    return  weights[indxmax]/wtSum < pcut


# normality test: is the MC integral normally distributed, with a small standard deviation?
#    - value: tests for converged integral
#    - arguments: 
#         - ncopies:               # of sub-integrals
#         - pcutNormalTest     Threshold p-value for normality test
#         - sigmaCutErrorThreshold   Threshold relative error in the integral
#    - implement normality test on **log(integral)** since the log should also be normally distributed if well converged
#           - this helps us handle large orders-of-magnitude differences
#           - compatible with a *relative* error threshold on integral
#           - only works for *positive-definite* integrands
#    - other python normality tests:  
#          scipy.stats.shapiro
#          scipy.stats.anderson
#  WARNING:
#    - this test assumes *unsorted* past history: the 'ncopies' segments are assumed independent.
import scipy.stats as stats
def convergence_test_NormalSubIntegrals(ncopies, pcutNormalTest, sigmaCutRelativeErrorThreshold, rvs, params):
    weights = rvs["integrand"]* rvs["joint_prior"]/rvs["joint_s_prior"]  # rvs["weights"] # rvs["weights"] is *sorted* (side effect?), breaking test. Recalculated weights are not.  Use explicitly calculated weights until sorting effect identified
#    weights = weights /numpy.sum(weights)    # Keep original normalization, so the integral values printed to stdout have meaning relative to the overall integral value.  No change in code logic : this factor scales out (from the log, below)
    igrandValues = numpy.zeros(ncopies)
    len_part = int(len(weights)/ncopies)  # deprecated: np.floor->np.int
    for indx in numpy.arange(ncopies):
        igrandValues[indx] = numpy.log(numpy.mean(weights[indx*len_part:(indx+1)*len_part]))  # change to mean rather than sum, so sub-integrals have meaning
    igrandValues= numpy.sort(igrandValues)#[2:]                            # Sort.  Useful in reports 
    valTest = stats.normaltest(igrandValues)[1]                              # small value is implausible
    igrandSigma = (numpy.std(igrandValues))/numpy.sqrt(ncopies)   # variance in *overall* integral, estimated from variance of sub-integrals
    print(" Test values on distribution of log evidence:  (gaussianity p-value; standard deviation of ln evidence) ", valTest, igrandSigma)
    print(" Ln(evidence) sub-integral values, as used in tests  : ", igrandValues)
    return valTest> pcutNormalTest and igrandSigma < sigmaCutRelativeErrorThreshold   # Test on left returns a small value if implausible. Hence pcut ->0 becomes increasingly difficult (and requires statistical accidents). Test on right requires relative error in integral also to be small when pcut is small.   FIXME: Give these variables two different names
    



"""
Simple program illustrating the precomputation step and calls to the
likelihood function.
"""

__author__ = "Evan Ochsner <evano@gravity.phys.uwm.edu>"

from factored_likelihood import *

Niter = 5 # Number of times to call likelihood function
Tmax = 38. # max ref. time
Tmin = 36. # min ref. time
Dmax = 110. * 1.e6 * lal.LAL_PC_SI # max ref. time
Dmin = 90. * 1.e6 * lal.LAL_PC_SI # min ref. time

#
# Produce data with a coherent signal in H1, L1, V1
#
data_dict = {}
psd_dict = {}
analyticPSD_Q = True # For simplicity, using an analytic PSD
psd_dict['H1'] = lal.LIGOIPsd
psd_dict['L1'] = lal.LIGOIPsd
psd_dict['V1'] = lal.LIGOIPsd

Psig = ChooseWaveformParams(fmin = 10., radec=True, theta=1.2, phi=2.4,
        detector='H1', dist=100.*1.e6*lal.LAL_PC_SI)
df = findDeltaF(Psig)
Psig.deltaF = df
data_dict['H1'] = lsu.non_herm_hoff(Psig)
Psig.detector = 'L1'
data_dict['L1'] = lsu.non_herm_hoff(Psig)
Psig.detector = 'V1'
data_dict['V1'] = lsu.non_herm_hoff(Psig)

# Struct to hold template parameters
P = ChooseWaveformParams(fmin = 40., dist=100.*1.e6*lal.LAL_PC_SI, deltaF=df)
Lmax = 2 # sets which modes to include

#
# Perform the Precompute stage
#
rholms_intp, crossTerms, rholms = PrecomputeLikelihoodTerms(P, data_dict,
        psd_dict, Lmax, analyticPSD_Q)
print("Finished Precomputation...")

#
# Call the likelihood function for various extrinsic parameter values
#
for i in range(Niter):
    P.phi = 2. * np.pi * np.random.rand() # right ascension
    P.theta = np.pi * np.random.rand() # declination
    P.tref = np.random.rand() * (Tmax - Tmin) # ref. time
    P.phiref = 2. * np.pi * np.random.rand() # ref. orbital phase
    P.incl = np.pi * np.random.rand() # inclination
    P.psi = np.pi * np.random.rand() # polarization angle
    P.dist = np.random.rand() * (Dmax - Dmin) # luminosity distance

    lnL = FactoredLogLikelihood(P, rholms_intp, crossTerms, Lmax)

    print("For (RA, DEC, tref, phiref, incl, psi, dist) =")
    print("\t", P.phi, P.theta, P.tref, P.phiref, P.incl, P.psi, P.dist)
    print("\tlog likelihood is:", lnL)



#!usr/env/bin python
'''
GPR_Fit.py
python 2.7.13
Vera Delfavero
Date updated: 02/02/2019

The function of this script is to provide the function, "fit_gp" to
the PE code (utils_ConstructIntrinsicPosterior_GenericCoordinates.py)

That function will fit training data, using a sparse gaussian proces,
and return a function which takes only test data as an input, 
and returns only the estimated mean. 
(Because that's all the PE code uses)
This saves a lot of time, because calculating the covariance
of the test data is expensive.
This function is also a sparse gaussian proces, which should make
it at least an order of magnitude faster than the sklearn 
gaussian process code.

The basic algorithm for this can be found on page 88 of R&W's
book on Gaussian Processes and Machine Learning. 
'''
from __future__ import print_function, division, unicode_literals
######## Kernels ########

from scipy.sparse import csc_matrix, linalg as sla

def compact_kernel(Xp, Xq,
                    thetas,
                    white_noise,
                    ):
    '''
    compact_kernel: Return a sparse representation of the 
        compact kernel described on page 88 of R&W.
        Most of the functions which generate the sparse kernel
        involve non-sparse matricies.

    Inputs:
        Xp: Dense array containing input vector
        Xq: Dense array containing additional input vector
            Note: if Xq is given, the returned kernel is the
                transpose of the true kernel
        thetas: Hyperparameters
            thetas[0]: Scale length used to construct sparse matrix.
    Outputs:
        K: Sparse kernel representation (csc matrix)
    '''
    import numpy as np
    
    # If no Xq, make one
    if Xq is None:
        Xq = np.copy(Xp)

    # Copy Xp
    Xp = np.copy(Xp).astype(float)
    Xq = np.copy(Xq).astype(float)

    # Identify dimensionality
    ndim = len(thetas)
    j = np.floor(float(ndim)/2.0) + 2

    # Scale by thetas
    if len(thetas) >= 1:
        for i in range(ndim):
            Xp[:,i] = Xp[:,i]/thetas[i]
            Xq[:,i] = Xq[:,i]/thetas[i]

    # Compute the difference matrix
    x2 = ((np.ones((Xq.shape + (Xp.shape[0],)))*Xp.T).T)
    x1 = (np.ones((Xp.shape + (Xq.shape[0],)))*Xq.T)

    # Assumption: real X and real k
    r = np.linalg.norm(x2 - x1, axis = 1)

    # K = ((1 - r)^2)_+
    eta = (1.0 - r)
    eta[eta < 0.0] = 0.0
    #K = eta**2
    K = (eta**(j + 1))*((j + 1)*r + 1.0)

    if white_noise != 0.0:
        K += white_noise*np.eye(K.shape[0], K.shape[1])

    # Construct sparse matrix
    K = csc_matrix(K)

    return K

######## Fit Function ########

def GP_fit_function(x_train, y_train, thetas, white_noise,
                    kernel_function = compact_kernel,
                    **kwargs):
    '''
    Objective: return a function which can sample the distribution
        with only the test inputs as an input.
        This function only needs to return a mean.
    '''
    from sksparse.cholmod import cholesky
    ## Guarantee dimensionality ##
    ## Find initial kernel ##
    K = kernel_function(x_train, x_train, thetas, white_noise)
    # find the alphas using cholesky decomposition
    L_factor = cholesky(K)
    alpha = L_factor(y_train)

    ## Construct function ##
    def my_fit(x_sample):
        #print(len(x_sample))
        Kt = kernel_function(x_sample, x_train, thetas, white_noise = 0.0)
        y_mean = Kt.dot(alpha)
        return y_mean

    return my_fit
    
######## Application for PE code ########

def fit_gp(x_train, y_train, y_errors=None,
        sparse_coeff = 3.0,
        white_noise = 0.01
        ):
    """
    This function can directly replace fit_gp in 
    utils_ConstructIntrinsicPosterior_GenericCoordinates.py

    It is an implimentation of my sparse fitting code,
    however most of this function was copied from the original.

    Inputs:
        x_train (numpy.ndarray): 
            Training data for model
            shape: (m, n) where m is the number of points,
                and n is the dimensionality
        y_train (numpy.ndarray): 
            Training data for model
            shape: (m, ) where m is the number of points
        sparse_coeff (float):
            This hyperparameter determines the sparseness of the matrix.
            A sparse_coeff of 1. is not sparse.
            A sparse_coeff of inf is diagonal.
            What value you can use for this is largely dependent on how
                much training data you have, as well as the dimensionality
                of your data
            About each dimension in parameter space, the scale length is
                calculated for that dimension as h/hs, where h is the range
                of the training data in that dimension.
        white_noise (float):
            The white noise hyperparemeter
    Outputs:
        my_fit (function(x_sample)):
            A function which will sample our model on x_sample,
                for an x_sample like x_train
    """
    import numpy as np
    # Find the dimensionality of the data from the shape of x_train
    n_dim = len(x_train[0])
    # Set the scale for each dimension
    h = np.empty(n_dim)
    for i in range(n_dim):
        h[i] = np.max(x_train) - np.min(x_train)
    thetas = h/sparse_coeff

    print(" GP: Input sample size ", x_train.shape)
    print(" Sparse scale lengths: ", thetas)

    ## Create function ##
    my_fit = GP_fit_function(x_train, y_train, thetas, white_noise)
    print(" std: ", np.std(my_fit(x_train) - y_train))

    return my_fit



#! /usr/bin/env python

# KJ Wagner 2023

import numpy as np
import argparse
#import os
#import sys
import pandas as pd
import configparser
import lal
import lalsimulation as lalsim
#from RIFT.misc.dag_utils import mkdir
import RIFT.lalsimutils as lalsimutils
#from RIFT.misc.dag_utils import which
#lalapps_path2cache = which('lal_path2cache')
from ligo.lw import lsctables, table, utils
#from scipy.stats import norm

parser = argparse.ArgumentParser()
parser.add_argument("--param-file",default=None,required=True,help='Population file with params to pass to PE')
parser.add_argument("--ini-file",default=None,required=True,help="read standard ini file so inj and PE match")
opts=  parser.parse_args()

param_file = opts.param_file

config = configparser.ConfigParser(allow_no_value=True)
config.read(opts.ini_file)

# Fixed time/sky location
fix_sky_location=config.getboolean('priors','fix_sky_location')
fix_event_time=config.getboolean('priors','fix_event_time')
fiducial_ra=float(config.get('priors','fiducial_ra'))
fiducial_dec=float(config.get('priors','fiducial_dec'))
fiducial_event_time=float(config.get('priors','fiducial_event_time'))

# Spin and eccentricity requests
no_spin=config.getboolean('priors','no_spin')
precessing_spin=config.getboolean('priors','precessing_spin')
aligned_spin=config.getboolean('priors','aligned_spin')
use_eccentric=config.getboolean('priors','use_eccentric')

# Noise/data requirements
ifos = ['H1','L1', 'V1']
channels = {'H1': 'FAKE-STRAIN', 'L1':'FAKE-STRAIN','V1':'FAKE-STRAIN'}
flow = {'H1':20, 'L1':20, 'V1':20}

fmin_template = float(config.get('waveform','fmin_template'))
fref = float(config.get('waveform','fmin_template')) # the same for now
fmax = float(config.get('data','fmax'))
srate_data = float(config.get('data', 'srate_data'))
seglen_data = float(config.get('data','seglen_data'))
seglen_analysis = float(config.get('data','seglen_analysis'))

# Approximant requests - TaylorF2Ecc requires extra (check rift_O4c for f_ecc???)
approx_str = config.get('waveform','approx')
if 'TaylorF2Ecc' in approx_str:
  set_fecc=20.0
  amporder=-1
lmax = 4

## Read in and store values from Zee
read_vals = pd.read_csv(param_file, header=0, index_col=False, delim_whitespace=True)
n_events = len(read_vals)

## Create list of all waveforms to send to xml, like ish pp_RIFT
# mass_1_source mass_2_source a_1 a_2 cos_tilt_1 cos_tilt_2 phi_1 phi_2 phi_12 eccentricity mean_anomaly redshift ra dec detection_time cos_iota psi phi_orb mass_1_detector mass_2_detector luminosity_distance
P_list =[]; indx=0
while len(P_list) < n_events:
  print("event ++{}".format(indx))
  P = lalsimutils.ChooseWaveformParams()
  # Randomize (sky location, etc)
  #P.randomize(dMax=d_max,dMin=d_min,aligned_spin_Q=aligned_spin,volumetric_spin_prior_Q=volumetric_spin,sMax=chi_max)
  P.tref = fiducial_event_time
  P.fmin = fmin_template
  P.fref = fref
  P.deltaF = 1./seglen_data
  P.deltaT = 1./srate_data

  # sky location
  if fix_sky_location:
      P.theta = fiducial_dec
      P.phi = fiducial_ra
  else:
      P.theta = read_vals['ra'][indx]
      P.phi = read_vals['dec'][indx]

  # spins, none and aligned for now
  if no_spin:
      P.s1x=P.s1y=P.s1z=0.0
      P.s2x=P.s2y=P.s2z=0.0
  elif aligned_spin:
      P.s1x = P.s1y=0
      P.s2x = P.s2y=0
      P.s1z = read_vals['spin1_z'][indx]
      P.s2z = read_vals['spin2_z'][indx]
  P.approx = lalsim.GetApproximantFromString(approx_str)
  
  # Read in values from Zee
  P.m1 = read_vals['mass_1_source'][indx] * lal.MSUN_SI
  P.m2 = read_vals['mass_2_source'][indx] * lal.MSUN_SI

  if use_eccentric:
    P.eccentricity = read_vals['eccentricity'][indx]
    P.meanPerAno = read_vals['mean_anomaly'][indx]
  if 'TaylorF2Ecc' in approx_str:
    P.ampO = amporder

  P.dist = read_vals['luminosity_distance'][indx] * 1e6 * lal.PC_SI
  P.phiref = read_vals['phi_orb'][indx]
  P.psi = read_vals['psi'][indx]
  P.incl = np.arccos(read_vals['cos_iota'][indx]) # check!!
  ## event time set in param file, but for inj set it??
  #P.tref = read_vals['detection_time'][indx] 
  
  P_list.append(P)
  indx+=1

lalsimutils.ChooseWaveformParams_array_to_xml(P_list,"mdc",fref=fmin_template, deltaF=1/16.)



#! /usr/bin/env python
#
# util_NRBestOfIndex.py
#
# GOAL
#    Accepts ILE default .dat output (indx m1 m2 group param XXX).
#    Finds the *single best* case in that set.
#    ---> FIXME, SHOULD USE LOCAL QUADRATIC FIT <--
#
#
#   python ~/research-projects/MonteCarloMarginalizeCode/Code/util_NRBestOfIndex.py --fname RIT-Generic-Lmax2-vC2.indexed --group Sequence-RIT-Generic


import argparse
import sys
import numpy as np
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import lalframe
import lal
import functools
import RIFT.interpolators.gp as gp

from multiprocessing import Pool
try:
    import os
    n_threads = int(os.environ['OMP_NUM_THREADS'])
    print(" Pool size : ", n_threads)
except:
    n_threads=1
    print(" - No multiprocessing - ")

try:
    import NRWaveformCatalogManager3 as nrwf
    hasNR =True
except:
    hasNR=False

###
### Load options
###

parser = argparse.ArgumentParser()
#parser.add_argument("--group",default=None)
parser.add_argument("--fname", default=None, help="Base output file for ascii text (.dat) and xml (.xml.gz)")
parser.add_argument("--verbose", action="store_true",default=False, help="Required to build post-frame-generating sanity-test plots")
parser.add_argument("--M-max-cut",type=float,default=1e5,help="Maximum mass to consider (e.g., if there is a cut on distance, this matters)")
parser.add_argument("--lnL-cut-up",default=5,type=float,help="Maximum amount (in lnL) a fit can increase lnL, based on what is calculated. Use a tighter threshold for a denser grid, and a wide threshold if you want to live dangerously.")
parser.add_argument("--sigma-cut",type=float,default=0.95,help="Eliminate points with large error from the fit.")
parser.add_argument("--fit", action="store_true",default=False, help="Local quadratic fit on best points")
parser.add_argument("--no-gp",action="store_true")
opts=  parser.parse_args()

if opts.verbose:
    True
    #lalsimutils.rosDebugMessagesContainer[0]=True   # enable error logging inside lalsimutils

if not opts.fname:
    sys.exit(0)



best_matches = {}
best_matches_masses ={}
full_spoke = {}

with open(opts.fname) as f:
 for line_str in f:
     # string split
     line = line_str.split()
#     if opts.verbose:
#         print "  Input ", line
     # Get NR data
     if float(line[1])+float(line[2]) > opts.M_max_cut:
         continue
     group = line[3]
     if not nrwf.internal_ParametersAreExpressions.has_key(group):
         continue
     param = None
     if nrwf.internal_ParametersAreExpressions[group]:
         param = eval(line[4])
     else:
         param = line[4]

     if len(line)<6:
         continue

     key = (group,param)

     failure_mode=False; sigma_here = 1e6
     # in case line abbreviated for some reason
     try:
         lnLhere = float(line[5])  
         sigma_here = float(line[6])
         npts_here = float(line[7])
     except:
         lnLhere = -10
         sigma_here = 0.1
         npts_here = 1
         failure_mode=True
         continue
     if sigma_here > opts.sigma_cut:
         failure_mode=True
         continue  # DO NOT RECORD ITEMS which are completely unconverged (one point). Insane answers likely.  (should ALSO have tunable cutoff on accuracy)
     if not failure_mode:
      if best_matches.has_key(key):
         if best_matches[key] < lnLhere:
                best_matches[key] = lnLhere
                best_matches_masses[key] = (float(line[1]),float(line[2]))
      else:
         best_matches[key] = line[5]
         best_matches_masses[key] = (float(line[1]),float(line[2]))

      # Add to spoke
      mtot = float(line[1])+ float(line[2])
      if full_spoke.has_key(key):
        full_spoke[key].append([mtot, lnLhere,sigma_here])
      else:
         full_spoke[key] = [[mtot,lnLhere, sigma_here]]

print(" -----  BEST SINGLE POINT MATCHES ------ ")  # unsorted
if  opts.fit:
  print("  +++ Using fit code +++ ")
  for key in best_matches:
      full_spoke[key] = np.array(sorted(np.array(full_spoke[key]),key=lambda p: p[1]))
      lnLmaxHere = np.max(full_spoke[key][:,1])
      indx_peak = np.argmax(full_spoke[key][:,1])
      sigma_crit = full_spoke[key][indx_peak][2]
      n_required =np.max([4,np.min([10, 0.1*len(full_spoke[key])])])
      indx_crit =int(np.max([n_required-1,np.argmin( np.abs(full_spoke[key][:,1]-10))]))
      if opts.verbose:
          print(" Spoke needs ", key, indx_crit+1)
      reduced_spoke = full_spoke[key][-indx_crit:]
      mMin = np.min(reduced_spoke[:,0])
      mMax = np.max(reduced_spoke[:,0])
      weights = reduced_spoke[:,2]
      reduced_spoke[np.isnan(weights),2] = 0  # do not use NAN entries with errors
      reduced_spoke[np.isnan(weights),1] = 0  # do not use NAN entries with errors
#      print " Fitting ", key, reduced_spoke
      z=[]; mBestGuess = 0; lnLBestGuess = 0;
      try:
          z = np.polyfit(reduced_spoke[:,0], reduced_spoke[:,1],2,w=1./(reduced_spoke[:,2])) # Note TERRIBLE documentation/convention in numpy https://docs.scipy.org/doc/numpy/reference/generated/numpy.polyfit.html
          mBestGuess = -0.5*z[1]/z[0]
          lnLBestGuess = z[2] -0.25*z[1]**2/z[0] 
      except:
          print(" Interpolation failure (internal to polyfit, VERY UNUSUAL) for spoke ", key, " reverting to pointwise best ")
          indxMax = np.argmax(reduced_spoke[:,2])
          mBestGuess = reduced_spoke[indxMax,0]
          lnLBestGuess = reduced_spoke[indxMax,2]
          z = [0,0,lnLBestGuess]
      print(key, z[0], mBestGuess, lnLBestGuess, best_matches[key], sigma_crit)
      if z[2]<0 and mBestGuess> mMin and mBestGuess < mMax:
        if lnLBestGuess < lnLmaxHere+opts.lnL_cut_up and lnLBestGuess > lnLmaxHere-5*sigma_crit:  # do not allow arbitrary extrapolation
          if opts.verbose:
              print(" Replacing peak ", key, best_matches[key], " -> ", lnLBestGuess, " at mass ", mBestGuess)
              if lnLBestGuess > lnLmaxHere+5:
                  print(" VERY LARGE CHANGE FOR", key, lnLmaxHere, "->", lnLBestGuess)
          best_matches[key] = lnLBestGuess 
          orig_mtot = best_matches_masses[key][0]+best_matches_masses[key][1]
          orig_m1 = best_matches_masses[key][0]
          orig_m2 = best_matches_masses[key][1]
          best_matches_masses[key]=(mBestGuess* orig_m1/orig_mtot, mBestGuess* orig_m2/orig_mtot)
        else:
          print(" Replacement rejected as out of range ", key, " reject ", lnLmaxHere, "->", lnLBestGuess, " : you probably need to rerun this spoke")
      if z[2]<0 and not ( mBestGuess> mMin and mBestGuess < mMax):
          print(" PLACEMENT FAILURE: ", key, mBestGuess, " outside of ", [mMin,mMax])
      

      ## REVISE BEST FIT VIA GAUSSIAN PROCESS
      #   GP is much less problematic for very crazy values that are far above trend due to bad MC luck
      #   Has VERY significant implications for rankings overall - often downranking
      try:
       if not opts.no_gp:
          my_gp = gp.GaussianProcess1d(reduced_spoke, sigma0=0.1,sigmab=0.2,h=1)
          xvals_dense = np.linspace(mMin,mMax,1000)
          yvals_dense = my_gp.predict(xvals_dense)[:,0]
          indx_gp_best = np.argmax(yvals_dense)
          print(" GP best fit ", xvals_dense[indx_gp_best], yvals_dense[indx_gp_best][0,0], "versus", mBestGuess, lnLmaxHere)
          if  mBestGuess-2 < xvals_dense[indx_gp_best] < mBestGuess+2 and yvals_dense[indx_gp_best][0,0] < lnLBestGuess+5*sigma_crit:
              mBestGuess = xvals_dense[indx_gp_best]
              lnLmaxHere= yvals_dense[indx_gp_best][0,0]
              best_matches[key] = lnLmaxHere
              orig_mtot = best_matches_masses[key][0]+best_matches_masses[key][1]
              orig_m1 = best_matches_masses[key][0]
              orig_m2 = best_matches_masses[key][1]
              best_matches_masses[key]=(mBestGuess* orig_m1/orig_mtot, mBestGuess* orig_m2/orig_mtot)
              print("  ....using GP best fit  for ", key, best_matches[key], best_matches_masses[key])

      except:
          print(" GP failure")


for key in best_matches:
  try:
    tmax =0
    if nrwf.internal_EstimatePeakL2M2Emission[key[0]].has_key(key[1]):
        tmax = nrwf.internal_EstimatePeakL2M2Emission[key[0]][key[1]]
    af = -1000
    Mf = -1000
    if nrwf.internal_WaveformMetadata[key[0]][key[1]].has_key('ChiFMagnitude'):
        af = nrwf.internal_WaveformMetadata[key[0]][key[1]]['ChiFMagnitude']
    if nrwf.internal_WaveformMetadata[key[0]][key[1]].has_key('MF'):
        Mf = nrwf.internal_WaveformMetadata[key[0]][key[1]]['MF']
    wfP = nrwf.WaveformModeCatalog(key[0],key[1], metadata_only=True)
    xi = wfP.P.extract_param('xi')
    print(best_matches[key], key[0], str(key[1]).replace(' ',''),   best_matches_masses[key][0], best_matches_masses[key][1], tmax, xi, Mf, af)
  except:
     print("Skipping ", key)



#! /usr/bin/env python
#
# ILE_data_handling_probe.py
#
# Like ILE, but only providing the Qlm and inverse spectrum filters
#

from __future__ import print_function

import sys
import functools
from optparse import OptionParser, OptionGroup

import matplotlib
matplotlib.use('agg')
from matplotlib import pyplot as plt

import numpy
import numpy as np
try:
  import cupy
  xpy_default=cupy
  identity_convert = cupy.asnumpy
except:
  print(' no cupy')
#  import numpy as cupy  # will automatically replace cupy calls with numpy!
  xpy_default=numpy  # just in case, to make replacement clear and to enable override
  identity_convert = lambda x: x  # trivial return itself


import lal
from glue.ligolw import utils, lsctables, table, ligolw
from glue.ligolw.utils import process
import glue.lal

import lalsimutils
import factored_likelihood
import RIFT.integrators.mcsampler as mcsampler
import xmlutils

#from lalinference.bayestar import fits as bfits

__author__ = "Evan Ochsner <evano@gravity.phys.uwm.edu>, Chris Pankow <pankow@gravity.phys.uwm.edu>, R. O'Shaughnessy"

#
# Pinnable parameters -- for command line processing
#
LIKELIHOOD_PINNABLE_PARAMS = ["right_ascension", "declination", "psi", "distance", "phi_orb", "t_ref", "inclination"]

def get_pinned_params(opts):
    """
    Retrieve a dictionary of user pinned parameters and their pin values.
    """
    return dict([(p,v) for p, v in opts.__dict__.items() if p in LIKELIHOOD_PINNABLE_PARAMS and v is not None]) 

def get_unpinned_params(opts, params):
    """
    Retrieve a set of unpinned parameters.
    """
    return params - set([p for p, v in opts.__dict__.items() if p in LIKELIHOOD_PINNABLE_PARAMS and v is not None])

#
# Option parsing
#

optp = OptionParser()
optp.add_option("-c", "--cache-file", default=None, help="LIGO cache file containing all data needed.")
optp.add_option("-C", "--channel-name", action="append", help="instrument=channel-name, e.g. H1=FAKE-STRAIN. Can be given multiple times for different instruments.")
optp.add_option("-p", "--psd-file", action="append", help="instrument=psd-file, e.g. H1=H1_PSD.xml.gz. Can be given multiple times for different instruments.")
optp.add_option("-k", "--skymap-file", help="Use skymap stored in given FITS file.")
optp.add_option("-x", "--coinc-xml", help="gstlal_inspiral XML file containing coincidence information.")
optp.add_option("-I", "--sim-xml", help="XML file containing mass grid to be evaluated")
optp.add_option("-E", "--event", default=0,type=int, help="Event number used for this run")
optp.add_option("--n-events-to-analyze", default=1,type=int, help="Number of events to analyze from this XML")
optp.add_option("--soft-fail-event-range",action='store_true',help='Soft failure (exit 0) if event ID is out of range. This happens in pipelines, if we have pre-built a DAG attempting to analyze more points than we really have')
optp.add_option("-f", "--reference-freq", type=float, default=100.0, help="Waveform reference frequency. Required, default is 100 Hz.")
optp.add_option("--fmin-template", dest='fmin_template', type=float, default=40, help="Waveform starting frequency.  Default is 40 Hz. Also equal to starting frequency for integration") 
optp.add_option("--fmin-ifo", action='append' , help="Minimum frequency for each IFO. Implemented by setting the PSD=0 below this cutoff. Use with care.") 
#optp.add_option("--nr-params",default=None, help="List of specific NR parameters and groups (and masses?) to use for the grid.")
#optp.add_option("--nr-index",type=int,default=-1,help="Index of specific NR simulation to use [integer]. Mass used: mtot= m1+m2")
optp.add_option('--nr-group', default=None,help="If using a *ssingle specific simulation* specified on the command line, provide it here")
optp.add_option('--nr-param', default=None,help="If using a *ssingle specific simulation* specified on the command line, provide it here")
optp.add_option("--nr-lookup",action='store_true', help=" Look up parameters from an NR catalog, instead of using the approximant specified")
optp.add_option("--nr-lookup-group",action='append', help="Restriction on 'group' for NR lookup")
optp.add_option("--nr-hybrid-use",action='store_true',help="Enable use of NR (or ROM!) hybrid, using --approx as the default approximant and with a frequency fmin")
optp.add_option("--nr-hybrid-method",default="taper_add",help="Hybridization method for NR (or ROM!).  Passed through to LALHybrid. pseudo_aligned_from22 will provide ad-hoc higher modes, if the early-time hybridization model only includes the 22 mode")
optp.add_option("--rom-group",default=None)
optp.add_option("--rom-param",default=None)
optp.add_option("--rom-use-basis",default=False,action='store_true',help="Use the ROM basis for inner products.")
optp.add_option("--rom-limit-basis-size-to",default=None,type=int)
optp.add_option("--rom-integrate-intrinsic",default=False,action='store_true',help='Integrate over intrinsic variables. REQUIRES rom_use_basis at present. ONLY integrates in mass ratio as present')
optp.add_option("--nr-perturbative-extraction",default=False,action='store_true')
optp.add_option("--nr-use-provided-strain",default=False,action='store_true')
optp.add_option("--no-memory",default=False,action='store_true', help="At present, turns off m=0 modes. Use with EXTREME caution only if requested by model developer")
optp.add_option("--restricted-mode-list-file",default=None,help="A list of ALL modes to use in likelihood. Incredibly dangerous. Only use when comparing with models which provide restricted mode sets, or otherwise to isolate the effect of subsets of modes on the whole")
optp.add_option("--use-external-EOB",default=False,action='store_true')
optp.add_option("--maximize-only",default=False, action='store_true',help="After integrating, attempts to find the single best fitting point")
optp.add_option("--dump-lnL-time-series",default=False, action='store_true',help="(requires --sim-xml) Dump lnL(t) at the injected parameters")
optp.add_option("-a", "--approximant", default="TaylorT4", help="Waveform family to use for templates. Any approximant implemented in LALSimulation is valid.")
optp.add_option("-A", "--amp-order", type=int, default=0, help="Include amplitude corrections in template waveforms up to this e.g. (e.g. 5 <==> 2.5PN), default is Newtonian order.")
optp.add_option("--l-max", type=int, default=2, help="Include all (l,m) modes with l less than or equal to this value.")
optp.add_option("-s", "--data-start-time", type=float, default=None, help="GPS start time of data segment. If given, must also give --data-end-time. If not given, sane start and end time will automatically be chosen.")
optp.add_option("-e", "--data-end-time", type=float, default=None, help="GPS end time of data segment. If given, must also give --data-start-time. If not given, sane start and end time will automatically be chosen.")
optp.add_option("--data-integration-window-half",default=50*1e-3,type=float,help="Only change this window size if you are an expert. The window for time integration is -/+ this quantity around the event time")
optp.add_option("-F", "--fmax", type=float, help="Upper frequency of signal integration. Default is use PSD's maximum frequency.")
optp.add_option("--srate",default=16384,type=int,help="Sampling rate. Change ONLY IF YOU ARE ABSOLUTELY SURE YOU KNOW WHAT YOU ARE DOING.")
optp.add_option("-t", "--event-time", type=float, help="GPS time of the event --- probably the end time. Required if --coinc-xml not given.")
optp.add_option("-i", "--inv-spec-trunc-time", type=float, default=8., help="Timescale of inverse spectrum truncation in seconds (Default is 8 - give 0 for no truncation)")
optp.add_option("-w", "--window-shape", type=float, default=0, help="Shape of Tukey window to apply to data (default is no windowing)")
optp.add_option("-m", "--time-marginalization", action="store_true", help="Perform marginalization over time via direct numerical integration. Default is false.")
optp.add_option("--vectorized", action="store_true", help="Perform manipulations of lm and timeseries using numpy arrays, not LAL data structures.  (Combine with --gpu to enable GPU use, where available)")
optp.add_option("--gpu", action="store_true", help="Perform manipulations of lm and timeseries using numpy arrays, CONVERTING TO GPU when available. You MUST use this option with --vectorized (otherwise it is a no-op). You MUST have a suitable version of cupy installed, your cuda operational, etc")
optp.add_option("--force-xpy", action="store_true", help="Use the xpy code path.  Use with --vectorized --gpu to use the fallback CPU-based code path. Useful for debugging.")
optp.add_option("-o", "--output-file", help="Save result to this file.")
optp.add_option("-O", "--output-format", default='xml', help="[xml|hdf5]")
optp.add_option("--verbose",action='store_true')
#
# Add the integration options
#
integration_params = OptionGroup(optp, "Integration Parameters", "Control the integration with these options.")
# Default is actually None, but that tells the integrator to go forever or until n_eff is hit.
integration_params.add_option("--interpolate-time", default=False,help="If using time marginalization, compute using a continuously-interpolated array. (Default=false)")
optp.add_option_group(integration_params)

#
# Add the intrinsic parameters
#
intrinsic_params = OptionGroup(optp, "Intrinsic Parameters", "Intrinsic parameters (e.g component mass) to use.")
#intrinsic_params.add_option("--pin-distance-to-sim",action='store_true', help="Pin *distance* value to sim entry. Used to enable source frame reconstruction with NR.")
intrinsic_params.add_option("--mass1", type=float, help="Value of first component mass, in solar masses. Required if not providing coinc tables.")
intrinsic_params.add_option("--mass2", type=float, help="Value of second component mass, in solar masses. Required if not providing coinc tables.")
intrinsic_params.add_option("--eff-lambda", type=float, help="Value of effective tidal parameter. Optional, ignored if not given.")
intrinsic_params.add_option("--deff-lambda", type=float, help="Value of second effective tidal parameter. Optional, ignored if not given")
optp.add_option_group(intrinsic_params)


#
# Add options to integrate over intrinsic parameters.  Same conventions as util_ManualOverlapGrid.py.  
# Parameters have special names, and we adopt priors that use those names.
# NOTE: Only 'q' implemented
#
intrinsic_int_params = OptionGroup(optp, "Intrinsic integrated parameters", "Intrinsic parameters to integrate over. ONLY currently used with ROM version")
intrinsic_int_params.add_option("--parameter",action='append')
intrinsic_int_params.add_option("--parameter-range",action='append',type=str)
intrinsic_int_params.add_option("--adapt-intrinsic",action='store_true')
optp.add_option_group(intrinsic_int_params)


#
# Add the pinnable parameters
#
pinnable = OptionGroup(optp, "Pinnable Parameters", "Specifying these command line options will pin the value of that parameter to the specified value with a probability of unity.")
for pin_param in LIKELIHOOD_PINNABLE_PARAMS:
    option = "--" + pin_param.replace("_", "-")
    pinnable.add_option(option, type=float, help="Pin the value of %s." % pin_param)
optp.add_option_group(pinnable)

opts, args = optp.parse_args()

if opts.gpu and xpy_default is numpy:
    print(" Override --gpu  (not available);  use --force-xpy to require the identical code path is used (with xpy =[np|cupy]")
    opts.gpu=False
    if opts.force_xpy:
        opts.gpu=True



deltaT = None
fSample= opts.srate # change sampling rate
if not(fSample is None):
  deltaT =1./fSample


# Load in restricted mode set, if available
restricted_mode_list=None
if not(opts.restricted_mode_list_file is None):
    modes =numpy.loadtxt(opts.restricted_mode_list_file,dtype=int) # columns are l m.  Must contain all. Only integers obviously
    restricted_mode_list = [ (l,m) for l,m in modes]
    print(" RESTRICTED MODE LIST target :", restricted_mode_list)

intrinsic_param_names = opts.parameter
valid_intrinsic_param_names = ['q']
if intrinsic_param_names:
 for param in intrinsic_param_names:
    
    # Check if in the valid list
    if not(param in valid_intrinsic_param_names):
            print(' Invalid param ', param, ' not in ', valid_intrinsic_param_names)
            sys.exit(0)
    param_ranges = []
    if len(intrinsic_param_names) == len(opts.parameter_range):
        param_ranges = numpy.array(map(eval, opts.parameter_range))
        # Rescale mass-dependent ranges to SI units
        for p in ['mc', 'm1', 'm2', 'mtot']:
          if p in intrinsic_param_names:
            indx = intrinsic_param_names.index(p)
            param_ranges[indx]= numpy.array(param_ranges[indx])* lal.MSUN_SI



# Check both or neither of --data-start/end-time given
if opts.data_start_time is None and opts.data_end_time is not None:
    raise ValueError("You must provide both or neither of --data-start-time and --data-end-time.")
if opts.data_end_time is None and opts.data_start_time is not None:
    raise ValueError("You must provide both or neither of --data-start-time and --data-end-time.")

#
# Import NR grid
#
NR_template_group=None
NR_template_param=None
if opts.nr_group and opts.nr_param:
    import NRWaveformCatalogManager3 as nrwf
    NR_template_group = opts.nr_group
    if nrwf.internal_ParametersAreExpressions[NR_template_group]:
        NR_template_param = eval(opts.nr_param)
    else:
        NR_template_param = opts.nr_param


#
# Hardcoded variables
#
template_min_freq = opts.fmin_template # minimum frequency of template
#t_ref_wind = 50e-3 # Interpolate in a window +/- this width about event time. 
t_ref_wind = opts.data_integration_window_half
T_safety = 2. # Safety buffer (in sec) for wraparound corruption

#
# Inverse spectrum truncation control
#
T_spec = opts.inv_spec_trunc_time
if T_spec == 0.: # Do not do inverse spectrum truncation
    inv_spec_trunc_Q = False
    T_safety += 8. # Add a bit more safety buffer in this case
else:
    inv_spec_trunc_Q = True



if opts.event_time is not None:
    event_time = glue.lal.LIGOTimeGPS(opts.event_time)
    print("Event time from command line: %s" % str(event_time))
else:
    print(" Error! ")
    sys.exit(0)

#
# Template descriptors
#

fiducial_epoch = lal.LIGOTimeGPS()
fiducial_epoch = event_time.seconds + 1e-9*event_time.nanoseconds   # no more direct access to gpsSeconds

# Struct to hold template parameters
P_list = None
if opts.sim_xml:
    print("====Loading injection XML:", opts.sim_xml, opts.event, " =======")
    P_list = lalsimutils.xml_to_ChooseWaveformParams_array(str(opts.sim_xml))
    if  len(P_list) < opts.event+opts.n_events_to_analyze:
        print(" Event list of range; soft exit")
        sys.exit(0)
    P_list = P_list[opts.event:(opts.event+opts.n_events_to_analyze)]
    for P in P_list:
        P.radec =False  # do NOT propagate the epoch later
        P.fref = opts.reference_freq
        P.fmin = template_min_freq
        P.tref = fiducial_epoch  # the XML table
        m1 = P.m1/lal.MSUN_SI
        m2 =P.m2/lal.MSUN_SI
        lambda1, lambda2 = P.lambda1,P.lambda2
        P.dist = factored_likelihood.distMpcRef * 1.e6 * lal.PC_SI   # use *nonstandard* distance
        P.phi=0.0
        P.psi=0.0
        P.incl = 0.0       # only works for aligned spins. Be careful.
        P.fref = opts.reference_freq
        if opts.approximant != "TaylorT4": # not default setting
            P.approx = lalsimutils.lalsim.GetApproximantFromString(opts.approximant)  # allow user to override the approx setting. Important for NR followup, where no approx set in sim_xml!

    P = P_list[0]  # Load in the physical parameters of the injection.  


# User requested bounds for data segment
if not (opts.data_start_time == None) and  not (opts.data_end_time == None):
    start_time =  opts.data_start_time
    end_time =  opts.data_end_time
    print("Fetching data segment with start=", start_time)
    print("                             end=", end_time)

# Automatically choose data segment bounds so region of interest isn't corrupted
# FIXME: Use estimate, instead of painful full waveform generation call here.
else:
    approxTmp = P.approx
    T_tmplt = lalsimutils.estimateWaveformDuration(P) + 4  # Much more robust than the previous (slow, prone-to-crash approach)
#     if (m1+m2)<30:
#         P.approx = lalsimutils.lalsim.TaylorT4  # should not impact length much.  IMPORTANT because EOB calls will fail at the default sampling rate
#         htmplt = lalsimutils.hoft(P)   # Horribly wasteful waveform gneeration solely to estimate duration.  Will also crash if spins are used.
#         P.approx = approxTmp
#         T_tmplt = - float(htmplt.epoch)    # ASSUMES time returned by hlm is a RELATIVE time that does not add tref. P.radec=False !
#         print fiducial_epoch, event_time, htmplt.epoch
#     else:
#         print " Using estimate for waveform length in a high mass regime: beware!"
#         T_tmplt = lalsimutils.estimateWaveformDuration(P) + 4
    T_seg = T_tmplt + T_spec + T_safety # Amount before and after event time
#    if opts.use_external_EOB:
#        T_seg *=2   # extra safety factor
    start_time = float(event_time) - T_seg
    end_time = float(event_time) + T_seg
    print("Fetching data segment with start=", start_time)
    print("                             end=", end_time)
    print("\t\tEvent time is: ", float(event_time))
    print("\t\tT_seg is: ", T_seg)

#
# Load in data and PSDs
#
data_dict, psd_dict = {}, {}

for inst, chan in map(lambda c: c.split("="), opts.channel_name):
    print("Reading channel %s from cache %s" % (inst+":"+chan, opts.cache_file))
    data_dict[inst] = lalsimutils.frame_data_to_non_herm_hoff(opts.cache_file,
            inst+":"+chan, start=start_time, stop=end_time,
            window_shape=opts.window_shape,deltaT=deltaT)
    print("Frequency binning: %f, length %d" % (data_dict[inst].deltaF,
            data_dict[inst].data.length))

flow_ifo_dict = {}
if opts.fmin_ifo:
 for inst, freq_str in map(lambda c: c.split("="), opts.fmin_ifo):
    freq_low_here = float(freq_str)
    print("Reading low frequency cutoff for instrument %s from %s" % (inst, freq_str), freq_low_here)
    flow_ifo_dict[inst] = freq_low_here

for inst, psdf in map(lambda c: c.split("="), opts.psd_file):
    print("Reading PSD for instrument %s from %s" % (inst, psdf))
    psd_dict[inst] = lalsimutils.get_psd_series_from_xmldoc(psdf, inst)

    deltaF = data_dict[inst].deltaF
    psd_dict[inst] = lalsimutils.resample_psd_series(psd_dict[inst], deltaF)
    print("PSD deltaF after interpolation %f" % psd_dict[inst].deltaF)

    # implement cutoff.  
    if inst in flow_ifo_dict.keys():
        if isinstance(psd_dict[inst], lal.REAL8FrequencySeries):
            psd_fvals = psd_dict[inst].f0 + deltaF*numpy.arange(psd_dict[inst].data.length)
            psd_dict[inst].data.data[ psd_fvals < flow_ifo_dict[inst]] = 0 # 
        else:
            print('FAIL')
            sys.exit(0)
#        elif isinstance(psd_dict[inst], pylal.xlal.datatypes.real8frequencyseries.REAL8FrequencySeries):  # for backward compatibility
#            psd_fvals = psd_dict[inst].f0 + deltaF*numpy.arange(len(psd_dict[inst].data))
#            psd_dict[inst].data[psd_fvals < ifo_dict[inst]] =0


    assert psd_dict[inst].deltaF == deltaF

    # Highest freq. at which PSD is defined
    # if isinstance(psd_dict[inst],
    #         pylal.xlal.datatypes.real8frequencyseries.REAL8FrequencySeries):
    #     fmax = psd_dict[inst].f0 + deltaF * (len(psd_dict[inst].data) - 1)
    if isinstance(psd_dict[inst], lal.REAL8FrequencySeries):
        fmax = psd_dict[inst].f0 + deltaF * (psd_dict[inst].data.length - 1)

    # Assert upper limit of IP integral does not go past where PSD defined
    assert opts.fmax is None or opts.fmax<= fmax
    # Allow us to target a smaller upper limit than provided by the PSD. Important for numerical PSDs that turn over at high frequency
    if opts.fmax and opts.fmax < fmax:
        fmax = opts.fmax # fmax is now the upper freq. of IP integral

# Ensure data and PSDs keyed to same detectors
if sorted(psd_dict.keys()) != sorted(data_dict.keys()):
    print("Got a different set of instruments based on data and PSDs provided.", file=sys.stderr)

# Ensure waveform has same sample rate, padded length as data
#
# N.B. This assumes all detector data has same sample rate, length
#
# data_dict holds 2-sided FrequencySeries, so their length is the same as
# that of the original TimeSeries that was FFT'd = Nsamples
# Also, deltaF = 1/T, with T = the duration (in sec) of the original TimeSeries
# Therefore 1/(data.length*deltaF) = T/Nsamples = deltaT
P.deltaT = 1./ (data_dict[data_dict.keys()[0]].data.length * deltaF)
P.deltaF = deltaF
for Psig in P_list:
  Psig.deltaT = P.deltaT
  Psig.deltaf = P.deltaF






def analyze_event(P_list, indx_event, data_dict, psd_dict, fmax, opts,inv_spec_trunc_Q=inv_spec_trunc_Q, T_spec=T_spec):
    nEvals=0
    P = P_list[indx_event]
    # Precompute
    t_window = 0.15
    rholms_intp, cross_terms, cross_terms_V,  rholms, rest=factored_likelihood.PrecomputeLikelihoodTerms(
            fiducial_epoch, t_window, P, data_dict, psd_dict, opts.l_max, fmax,
            False, inv_spec_trunc_Q, T_spec,
            NR_group=NR_template_group,NR_param=NR_template_param,
            use_external_EOB=opts.use_external_EOB,nr_lookup=opts.nr_lookup,nr_lookup_valid_groups=opts.nr_lookup_group,perturbative_extraction=opts.nr_perturbative_extraction,use_provided_strain=opts.nr_use_provided_strain,hybrid_use=opts.nr_hybrid_use,hybrid_method=opts.nr_hybrid_method,ROM_group=opts.rom_group,ROM_param=opts.rom_param,ROM_use_basis=opts.rom_use_basis,verbose=opts.verbose,quiet=not opts.verbose,ROM_limit_basis_size=opts.rom_limit_basis_size_to,no_memory=opts.no_memory,skip_interpolation=opts.vectorized)


    # Only allocate the FFT plans ONCE
    ifo0 = psd_dict.keys()[0]
    npts = data_dict[ifo0].data.length
    print(" Allocating FFT forward, reverse plans ", npts)
    fwdplan = lal.CreateForwardREAL8FFTPlan(npts, 0)
    revplan = lal.CreateReverseREAL8FFTPlan(npts, 0)

    for ifo in psd_dict:
        # Plot PSDs
        plt.figure(99) # PSD figure)
        fvals = psd_dict[ifo].f0 + np.arange(psd_dict[ifo].data.length)*psd_dict[ifo].deltaF
        plt.plot(fvals,np.log10(np.sqrt(psd_dict[ifo].data.data)),label=ifo)
        plt.figure(88)
        plt.plot(fvals, fvals**(-14./6.) /psd_dict[ifo].data.data)
        plt.figure(1)

        # Plot inverse spectrum filters
        #     - see lalsimutils code for inv_spec_trunc_Q , copied verbatim
        plt.figure(98)
        IP_here = lalsimutils.ComplexOverlap(P.fmin, fmax, 1./2/P.deltaT, P.deltaF, psd_dict[ifo], False,
                                             inv_spec_trunc_Q, T_spec)
        WFD = lal.CreateCOMPLEX16FrequencySeries('FD root inv. spec.',
                                                 lal.LIGOTimeGPS(0.), 0., IP_here.deltaF,
                                                 lal.DimensionlessUnit, IP_here.len1side)
        WTD = lal.CreateREAL8TimeSeries('TD root inv. spec.',
                    lal.LIGOTimeGPS(0.), 0., IP_here.deltaT,
                    lal.DimensionlessUnit, IP_here.len2side)
        print(ifo, IP_here.len2side)
#        if not(fwdplan is None):
        WFD.data.data[:] = np.sqrt(IP_here.weights) # W_FD is 1/sqrt(S_n(f))
        WFD.data.data[0] = WFD.data.data[-1] = 0. # zero 0, f_Nyq bins
        lal.REAL8FreqTimeFFT(WTD, WFD, revplan) # IFFT to TD
        tvals = IP_here.deltaT*np.arange(IP_here.len2side)
        plt.plot(tvals, np.log10(np.abs(WTD.data.data)),label=ifo)

        # Plot Q's
        plt.figure(1);
        plt.clf()
        for mode in rholms[ifo]:
            Q_here = rholms[ifo][mode]
            tvals = lalsimutils.evaluate_tvals(Q_here) - P.tref
            plt.plot(tvals, np.abs(Q_here.data.data),label=ifo+str(mode[0])+","+str(mode[1]))
        plt.legend();
        plt.xlabel("t (s)"); plt.xlim(-0.1,0.1); plt.xlabel("$Q_{lm}$")
        plt.title(ifo)
        plt.savefig(ifo+ "_Q.png")
        
    plt.figure(99); plt.legend(); plt.xlabel("f(Hz)"); plt.ylabel(r'$\sqrt{S_h}$'); plt.ylim(-24,-21)
    plt.savefig("PSD_plots.png")
    plt.xlim(5,800)
    plt.savefig("PSD_plots_detail.png")
    plt.xlim(10,400); plt.ylim(-23.5,-22)
    plt.savefig("PSD_plots_detail_fine.png")
    plt.figure(98); plt.legend(); plt.xlabel("t (s)")
    plt.savefig("PSD_inv_plots.png")
    plt.figure(88); plt.legend(); plt.xlabel("f (Hz)"); plt.ylabel(r'$(S_h f^{15/6})^{-1}$')
    plt.savefig("PSD_weight_f.png")

lnL_sofar = -np.inf
no_adapt_sky = False
for indx in numpy.arange(len(P_list)):
  res = analyze_event(P_list, indx, data_dict, psd_dict, fmax, opts)



# Test for evaluating rosenbrock 2d likelihood, where the 1d marginal and evidence can be computed.
#
# Suggested:
#   python test_mcsampler_rosenbrock.py;  plot_posterior_corner.py --posterior-file fairdraw_rosenbrock_1.dat --posterior-file fairdraw_rosenbrock_1b.dat --posterior-file fairdraw_rosenbrock_2.dat --parameter x1 --parameter x2  --quantiles None --ci-list [0.9]

from __future__ import print_function
import numpy as np
from scipy.stats import multivariate_normal
from scipy.stats import truncnorm
import matplotlib.pyplot as plt

from scipy.special import erf

from RIFT.integrators import mcsampler, mcsamplerEnsemble, mcsamplerGPU, mcsamplerAdaptiveVolume

import optparse
parser = optparse.OptionParser()
parser.add_option("--n-max",type=int,default=1000000)
parser.add_option("--save-plot",action='store_true')
parser.add_option("--production",action='store_true')
parser.add_option("--use-lnL",action='store_true')
parser.add_option("--lnL-shift",default=100,type=float,help="Our integrators assume lnL >0 when designing adaptation, so this shift helps us adapt, and is more consistent with our real problems.  Choose adapt-weight-exponent consistent with this value to be most realistic")
#parser.add_option("--as-test",action='store_true')
#parser.add_option("--no-adapt",action='store_true')
parser.add_option("--floor-level",default=0.4,type=float)  # for this problem, a higher floor level helps
parser.add_option("--adapt-weight-exponent",default=0.1,type=float)
parser.add_option("--n-chunk",default=10000,type=int)
parser.add_option("--verbose",action='store_true')
opts, args = parser.parse_args()


tempering_exp =opts.adapt_weight_exponent
# max number of samples for mcsampler
nmax = opts.n_max
n_block=opts.n_chunk
n_iters = nmax/n_block

save_fairdraws=True
save_fairdraw_prefix="fairdraw_rosenbrock"
#

Z_rosenbrock = -5.804

### test parameters

### define integrand 
### some typecasting needed
lnL_offset = opts.lnL_shift    # lnL > 1 used in adaptation
def f(x1, x2):
    minus_lnL = np.array(np.power((1.-x1), 2) + 100.* np.power((x2-x1**2),2),dtype=float)
    return np.exp( lnL_offset - (minus_lnL))
def ln_f(x1, x2): 
    minus_lnL = np.array(np.power((1.-x1), 2) + 100.* np.power((x2-x1**2),2),dtype=float)
    return lnL_offset - minus_lnL

### initialize samplers
sampler = mcsampler.MCSampler()
samplerEnsemble = mcsamplerEnsemble.MCSampler()
samplerAC = mcsamplerGPU.MCSampler()
samplerAV = mcsamplerAdaptiveVolume.MCSampler()

### add parameters
llim=-5
rlim=5
params = ['x1','x2']
for p in params:
    sampler.add_parameter(p, np.vectorize(lambda x:1), 
            prior_pdf=np.vectorize(lambda x:1./(rlim-llim)),
            left_limit=llim, right_limit=rlim,
            adaptive_sampling=True)
    samplerEnsemble.add_parameter(p, 
                                  pdf=np.vectorize(lambda x:1/(rlim-llim)),
                                  prior_pdf=np.vectorize(lambda x:1/(rlim-llim)),
                                  left_limit=llim, right_limit=rlim,adaptive_sampling=True)
    # for AC sampler, make sure pdf and prior pdfs are *normalized* *initially*
    samplerAC.add_parameter(p, pdf=np.vectorize(lambda x:1/(rlim-llim)),
            prior_pdf=np.vectorize(lambda x:1/(rlim-llim)),
            left_limit=llim, right_limit=rlim,adaptive_sampling=True)
    samplerAV.add_parameter(p, pdf=np.vectorize(lambda x:1/(rlim-llim)),
            prior_pdf=np.vectorize(lambda x:1/(rlim-llim)),
            left_limit=llim, right_limit=rlim,adaptive_sampling=True)

# number of Gaussian components to use in GMM
n_comp = 2

extra_args = {"n": opts.n_chunk,"n_adapt":100, "floor_level":opts.floor_level,"tempering_exp" :tempering_exp}


### integrate
integral_1, var_1, eff_samp_1, _ = sampler.integrate(f, *params, 
        no_protect_names=True, nmax=nmax, save_intg=True,verbose=opts.verbose,**extra_args)
print(np.log(integral_1) - lnL_offset, Z_rosenbrock)
print(" --- finished default --")
integral_1b, var_1b, eff_samp_1b, _ = samplerAC.integrate(f, *params, 
        no_protect_names=True, nmax=nmax, save_intg=True,verbose=opts.verbose,**extra_args)
print(np.log(integral_1b) - lnL_offset, Z_rosenbrock)
print(" --- finished AC --")
print(" NEED TO ADD OPTION TO TEST CORRELATED SAMPLING ")
use_lnL = opts.use_lnL
return_lnI=opts.use_lnL
if use_lnL:
    infunc = ln_f
else:
    infunc = f
integral_2, var_2, eff_samp_2, _ = samplerEnsemble.integrate(infunc, *params, 
        min_iter=n_iters, max_iter=n_iters, correlate_all_dims=True, n_comp=n_comp,super_verbose=False,verbose=opts.verbose,use_lnL=use_lnL,return_lnI=return_lnI,**extra_args)
if return_lnI and use_lnL:
    integral_2 = np.exp(integral_2)
print(np.log(integral_2) - lnL_offset, Z_rosenbrock)
print(" --- finished GMM --")
#def my_infunc(*args):
#    return ln_f(*np.array([*args]).T)
samplerAV.setup()
integral_3, var_3, eff_samp_3, _ = samplerAV.integrate_log(ln_f, *params, no_protect_names=True,
        nmax=nmax,
        min_iter=n_iters, max_iter=n_iters, **extra_args)
integral_3 = np.exp(integral_3)
var_3 = np.exp(var_3)



print(integral_1/np.exp(lnL_offset),integral_1b/np.exp(lnL_offset),integral_2/np.exp(lnL_offset),integral_3/np.exp(lnL_offset),np.exp(Z_rosenbrock))
print(np.array([integral_1,integral_1b,integral_2,integral_3])/np.exp(Z_rosenbrock+lnL_offset))
print(" AC/default ", integral_1b/integral_1, np.sqrt(var_1)/integral_1)  # off by width**3
print(" GMM/default ",integral_2/integral_1, np.sqrt(var_1)/integral_1, np.sqrt(var_2)/integral_2)
print(" AV/default ",integral_3/integral_1, np.sqrt(var_1)/integral_1, np.sqrt(var_3)/integral_3)
### CDFs

### get our posterior samples as a single array
ndim=2
arr_1 = np.empty((len(sampler._rvs["x1"]), ndim))
arr_1b = np.empty((len(samplerAC._rvs["x1"]), ndim))
arr_2 = np.empty((len(samplerEnsemble._rvs["x1"]), ndim))
arr_3 = np.empty((len(samplerAV._rvs["x1"]), ndim))
for i in range(ndim):
    arr_1[:,i] = sampler._rvs["x"+str(i+1)].flatten()
    arr_1b[:,i] = samplerAC._rvs["x"+str(i+1)].flatten()
    arr_2[:,i] = samplerEnsemble._rvs["x"+str(i+1)].flatten()
    arr_3[:,i] = samplerAV._rvs["x"+str(i+1)].flatten()

colors = ["black", "red", "blue", "green", "orange"]

plt.figure(figsize=(10, 8))

if True:
    # NOTE: old mcsampler stores L, mcsamplerEnsemble stores lnL
    L = sampler._rvs["integrand"]
    p = sampler._rvs["joint_prior"]
    ps = sampler._rvs["joint_s_prior"]
    ### compute weights of samples
    weights_1 = (L * p / ps)
    n_ess_1 = np.sum(weights_1)**2/np.sum(weights_1**2)
    print("default  n_eff, n_ess ", eff_samp_1,n_ess_1)

    L = samplerEnsemble._rvs["integrand"]
    if return_lnI:
        L = np.exp(L - np.max(L))
    p = samplerEnsemble._rvs["joint_prior"]
    ps = samplerEnsemble._rvs["joint_s_prior"]
    ### compute weights of samples
    weights_2 = (L * p / ps)
    n_ess_2 = np.sum(weights_2)**2/np.sum(weights_2**2)
    print("AC  n_eff, n_ess ", eff_samp_2,n_ess_2)


    L = samplerAC._rvs["integrand"]
    p = samplerAC._rvs["joint_prior"]
    ps = samplerAC._rvs["joint_s_prior"]
    ### compute weights of samples
    weights_1b = (L * p / ps)
    n_ess_1b = np.sum(weights_1b)**2/np.sum(weights_1b**2)
    print("GMM  n_eff, n_ess ", eff_samp_1b,n_ess_1b)

    lnL = samplerAV._rvs["log_integrand"]  
    lnp = samplerAV._rvs["log_joint_prior"]  
    lnps = samplerAV._rvs["log_joint_s_prior"]  
    ### compute weights of samples
    weights_3 = np.exp(lnL + lnp - lnps )
    n_ess_3 = np.sum(weights_3)**2/np.sum(weights_3**2)
    print("AV  n_eff, n_ess ", eff_samp_3,n_ess_3)


if save_fairdraws:
    npts_out_1 = int(n_ess_1)
    p = np.array(weights_1/np.sum(weights_1),dtype=np.float64)
    indx_save_1 = np.random.choice(np.arange(len(p)), size=npts_out_1, p=p)
    np.savetxt(save_fairdraw_prefix+"_1.dat", arr_1[indx_save_1],header=" x1 x2")

    npts_out_2 = int(n_ess_2)
    p = np.array(weights_2/np.sum(weights_2),dtype=np.float64)
    indx_save_2 = np.random.choice(np.arange(len(p)), size=npts_out_2, p=p)
    np.savetxt(save_fairdraw_prefix+"_2.dat", arr_2[indx_save_2],header=" x1 x2")

    npts_out_1b = int(n_ess_1b)
    p = np.array(weights_1b/np.sum(weights_1b),dtype=np.float64)
    indx_save_1b = np.random.choice(np.arange(len(p)), size=npts_out_1b, p=p)
    np.savetxt(save_fairdraw_prefix+"_1b.dat", arr_1b[indx_save_1b],header=" x1 x2")

    npts_out_3 = np.min([int(n_ess_3*0.8),5000])
    p = np.array(weights_3/np.sum(weights_3),dtype=np.float64)
    indx_save_3 = np.random.choice(np.arange(len(p)), size=npts_out_3, p=p)
    np.savetxt(save_fairdraw_prefix+"_3.dat", arr_3[indx_save_3],header=" x1 x2")


##
## HOW TO MAKE JS TEST
# for param in x1 x2 ; do ls fairdraw_*.dat | ../scripts/tool_pairs.py --prefix="convergence_test_samples.py --method JS --parameter ${param}" > tmpfile;./tmpfile > jsvals_${param}.dat; done

## HOW TO MAKE A PLOT
# plot_posterior_corner.py --posterior-file fairdraw_rosenbrock_1.dat --posterior-file fairdraw_rosenbrock_1b.dat --posterior-file fairdraw_rosenbrock_2.dat --parameter x1 --parameter x2  --quantiles None --ci-list [0.9]


# Make copies, so they have the same orders relative to the sample values
weights_1_orig = np.array(weights_1)
weights_1b_orig = np.array(weights_1b)
weights_2_orig = np.array(weights_2)
weights_3_orig = np.array(weights_3)
# Note: the fairdraw export is better
for i in range(ndim):
    ### get sorted samples (for the current dimension)
    indx_sort = np.argsort(arr_1[:,i])
    x_1 = arr_1[:,i][indx_sort]
    weights_1 = weights_1_orig[indx_sort]

    indx_sort = np.argsort(arr_1b[:,i])
    x_1b = arr_1b[:,i][indx_sort]
    weights_1b = weights_1b_orig[indx_sort]

    indx_sort = np.argsort(arr_2[:,i])
    x_2 = arr_2[:,i][indx_sort]
    weights_2 = weights_2_orig[indx_sort]

    indx_sort = np.argsort(arr_3[:,i])
    x_3 = arr_3[:,i][indx_sort]
    weights_3 = weights_3_orig[indx_sort]


    y_1 = np.cumsum(weights_1)
    y_1 /= y_1[-1] # normalize
    y_2 = np.cumsum(weights_2)
    y_2 /= y_2[-1] # normalize
    y_1b = np.cumsum(weights_1b)
    y_1b /= y_1b[-1]
    y_3 = np.cumsum(weights_3)
    y_3 /= y_3[-1]
    ### plot recovered cdf
    plt.plot(x_1, y_1, "--", label="CDF, mcsampler", color=colors[i],
            linewidth=2)
    plt.plot(x_2, y_2, ":", label="CDF, mcsamplerEnsemble",
            color=colors[i], linewidth=2)
    plt.plot(x_1b, y_1b, ".", label="CDF, mcsamplerAC",
            color=colors[i], linewidth=1)
    plt.plot(x_3, y_3, "--", label="CDF, mcsamplerAV",
            color=colors[i], linewidth=1)


# Semianalytic plot
def fn_marginal_x(x):
    return 1/(10) * np.sqrt(np.pi)/2000 * (erf( 10*(5-x**2)) + erf(10*(5+x**2))) * np.exp( - (1-x)**2)
def dP_cdf(p,x):
    return fn_marginal_x(x)

from scipy import integrate, interpolate

x_i = np.linspace(-5,5,1000)
cdf = integrate.odeint(dP_cdf, [0], x_i, hmax=0.01*(10)).T[0]
cdf/=cdf[-1]
plt.plot(x_i,cdf,label="true")



if not(opts.production):
    plt.legend()

fname = "cdf_rosenbrock.png"

print("Saving CDF figure as " + fname + "...")
if opts.production:
    plt.xlabel(r'$x$')
    plt.ylabel(r'$P(<x)$')

plt.savefig(fname)



"""
test_precompute.py:  Testing the likelihood evaluation.
    - Generates a fiducial signal
    - The synthetic signal is introduced directly overhead a single detector at the earth's barycenter.
      I assume *zero noise* and *an assumed known PSD*
    - Generates a 'template' signal at a reference distance
    - Precomputes all factors appearing in the likelihood
      Compares those factors for self-consistency (U and Q)
    - Evaluates the likelihood at the injection parameters and versus time

"""

__author__ = "R. O'Shaughnessy"

# GOAL
#   Like TestPrecompute, but using *extremely controlled* conditions so I can *calibrate a priori*
#   The source is directly overhead a single interferometer.  


from factored_likelihood import *
from matplotlib import pylab as plt
import sys

checkResults = True # Turn on to print/plot output; Turn off for testing speed
checkInputPlots = False
checkResultsPlots = True
checkResultsSlowChecks = False
rosUseRandomEventTime = True
rosUseDifferentWaveformLengths = False    # Very important test: lnL should be independent of the template length
rosUseRandomTemplateStartingFrequency = True

data_dict = {}
psd_dict = {}
rhoExpected ={}
rhoExpectedAlt ={}
analyticPSD_Q = True # For simplicity, using an analytic PSD

fminWavesSignal = 25
if rosUseDifferentWaveformLengths: 
    fminWavesTemplate = fminWavesSignal+0.005
else:
    if rosUseRandomTemplateStartingFrequency:
         print("   --- Generating a random template starting frequency  ---- ")
         fminWavesTemplate = fminWavesSignal+5.*np.random.random_sample()
    else:
        fminWavesTemplate = fminWavesSignal

fminSNR = 25
fSample = 4096*4

ifoName = "Fake"
theEpochFiducial = lal.LIGOTimeGPS(1064023405.000000000)   # 2013-09-24 early am 

# Create complex FD data that does not assume Hermitianity - i.e.
# contains positive and negative freq. content
def non_herm_hoff_fake(P):
    hp, hc = lalsim.SimInspiralChooseTDWaveform(P.phiref, P.deltaT, P.m1, P.m2, 
            P.s1x, P.s1y, P.s1z, P.s2x, P.s2y, P.s2z, P.fmin, P.fref, P.dist, 
            P.incl, P.lambda1, P.lambda2, P.waveFlags, P.nonGRparams,
            P.ampO, P.phaseO, P.approx)
    hp.epoch = hp.epoch + P.tref
    hc.epoch = hc.epoch + P.tref
    hoft = hp
    if P.taper != lalsim.LAL_SIM_INSPIRAL_TAPER_NONE: # Taper if requested
        lalsim.SimInspiralREAL8WaveTaper(hoft.data, P.taper)
    if P.deltaF == None:
        TDlen = nextPow2(hoft.data.length)
    else:
        TDlen = int(1./P.deltaF * 1./P.deltaT)
        assert TDlen >= hoft.data.length

    fwdplan=lal.CreateForwardCOMPLEX16FFTPlan(TDlen,0)
    hoft = lal.ResizeREAL8TimeSeries(hoft, 0, TDlen)
    hoftC = lal.CreateCOMPLEX16TimeSeries("hoft", hoft.epoch, hoft.f0,
            hoft.deltaT, hoft.sampleUnits, TDlen)
    # copy h(t) into a COMPLEX16 array which happens to be purely real
    for i in range(TDlen):
        hoftC.data.data[i] = hoft.data.data[i]
    FDlen = TDlen
    hoff = lal.CreateCOMPLEX16FrequencySeries("Template h(f)", 
            hoft.epoch, hoft.f0, 1./hoft.deltaT/TDlen, lal.lalHertzUnit, 
            FDlen)
    lal.COMPLEX16TimeFreqFFT(hoff, hoftC, fwdplan)
    return hoff


distanceFiducial = 25.  # Make same as reference
psd_dict[ifoName] =  lalsim.SimNoisePSDiLIGOSRD
m1 = 4*lal.LAL_MSUN_SI
m2 = 3*lal.LAL_MSUN_SI
tEventFiducial = 0.000 # 10./fSample
if rosUseRandomEventTime:
    print("   --- Generating a random event (barycenter) time  ---- ")
    tEventFiducial+= 0.05*np.random.random_sample()
ampO =0 # sets which modes to include in the physical signal
Lmax = 2  # sets which modes to include in the output
fref = 100
Psig = ChooseWaveformParams(fmin = fminWavesSignal, radec=False, incl=0.0,phiref=0.0, theta=0.0, phi=0,psi=0.0,
         m1=m1,m2=m2,
         ampO=ampO,
         fref=fref,
         tref=theEpochFiducial+tEventFiducial,
         deltaT=1./fSample,
         dist=distanceFiducial*1.e6*lal.LAL_PC_SI)
tEventFiducialGPS = Psig.tref             # the 'trigger time' we will localize on
df = findDeltaF(Psig)
Psig.deltaF = df
Psig.print_params()
data_dict[ifoName] = non_herm_hoff_fake(Psig)
print("Timing spacing in data vs expected : ", df, data_dict[ifoName].deltaF)

print(" == Data report == ")
detectors = data_dict.keys()
rho2Net = 0
fvalsDump = np.linspace(20,2000,1000)
psdVecDump = map(lambda x: psd_dict[ifoName](x),fvalsDump )
#print psdVecDump
np.savetxt('psd.dat', (fvalsDump,psdVecDump))  # to calibrate the SNR calculations below

for det in detectors:
    IP = ComplexIP(fLow=fminSNR, fNyq=fSample/2.,deltaF=df,psd=psd_dict[det])
    IPOverlap = ComplexOverlap(fLow=fminSNR, fNyq=fSample/2.,deltaF=df,psd=psd_dict[det],analyticPSD_Q=True,full_output=True)  # Use for debugging later
    rhoExpected[det] = rhoDet = IP.norm(data_dict[det])
    rhoExpectedAlt[det] = rhoDet2 = IPOverlap.norm(data_dict[det])
    rho2Net += rhoDet*rhoDet
    print(det, rhoDet, rhoDet2, " at epoch ", float(data_dict[det].epoch))
print("Network : ", np.sqrt(rho2Net))


print(" ======= Template specified: precomputing all quantities ==========")
# Struct to hold template parameters
# Fiducial distance provided but will not be used
P = ChooseWaveformParams(fmin=fminWavesTemplate, radec=False, incl=0.0,phiref=0.0, theta=0.0, phi=0,psi=0.0,
         m1=m1,m2=m2,
         ampO=ampO,
         fref=fref,
         tref=theEpochFiducial,
         deltaT=1./fSample,
         dist=100*1.e6*lal.LAL_PC_SI,
         deltaF=df) #ChooseWaveformParams(m1=m1,m2=m2,fmin = fminWaves, dist=100.*1.e6*lal.LAL_PC_SI, deltaF=df,ampO=ampO,fref=fref)
rholms_intp, crossTerms, rholms, epoch_post = PrecomputeLikelihoodTerms(theEpochFiducial,P, data_dict, psd_dict, Lmax, analyticPSD_Q)

if checkResults == True:
    # Print values of cross terms
    detectors = data_dict.keys()
    assert rholms_intp.keys() == crossTerms.keys()
    for det in detectors:
        for pair1 in rholms_intp[ifoName]:
            for pair2 in rholms_intp[ifoName]:
                if np.abs(crossTerms[det][pair1,pair2]) > 1e-5:
                    print(det, pair1, pair2, crossTerms[det][pair1,pair2], " compare (2,\pm 2) in scale to ", rhoExpected[det]**2 * 8.*lal.LAL_PI/5. *np.power( distanceFiducial/distMpcRef,2))
    
    print(" ======= UV symmetry check (reflection symmetric) ==========")
    constraint1 = 0
    for det in detectors:
        for pair1 in rholms_intp[ifoName]:
            for pair2 in rholms_intp[ifoName]:
#                print pair1, pair2, crossTerms[det][pair1,pair2], " - ",  ((-1)**(pair1[0]+pair2[0])*np.conj(crossTerms[det][(pair1[0],-pair1[1]),(pair2[0],-pair2[1])])
                constraint1 += np.abs( crossTerms[det][pair1,pair2] - ((-1)**pair1[0])*np.conj(crossTerms[det][(pair1[0],-pair1[1]), (pair2[0],-pair2[1])]) )**2
    print("   : Reflection symmetry constraint (UV) : 0 ~=  ", constraint1)

    print(" ======= UV test: Recover the SNR of the injection  ==========")
    print(" Detector lnLmodel  (-2lnLmodel)^(1/2)  rho(directly)  [last two entries should be equal! (if source and template start at the same frequency)] ")
    for det in detectors:
        lnLModel = SingleDetectorLogLikelihoodModel(crossTerms, Psig.tref, Psig.phi, Psig.theta, Psig.incl, Psig.phiref, Psig.psi, Psig.dist, 2, det)
        print(det, lnLModel, np.sqrt(-2*lnLModel), rhoExpected[det])


    print(" ======= rholm complex conjugation check (22 and 2-2 modes only) ==========")
    constraint1 = 0
    for det in detectors:
        hxx = lalsim.SphHarmTimeSeriesGetMode(rholms[det], 2, 2)
        hyy = lalsim.SphHarmTimeSeriesGetMode(rholms[det], 2, -2)
        for i in np.arange(len(hxx.data.data)):
            constraint1+= np.abs(hxx.data.data[i]-np.conj(hyy.data.data[i]))**2
    print("   : Reflection symmetry constraint (Q22,Q2-2) with raw data: : 0 ~= ", constraint1/len(hxx.data.data))    # error per point

    # constraint1 = 0
    # for det in detectors:
    #     npts = len(hxx.data.data)
    #     t= hxx.deltaT*np.arange(npts)
    #     for i in np.arange(len(hxx.data.data)):
    #         constraint1+= np.abs(rholms_intp[det][(2,2)](t[i])-np.conj(rholms_intp[det][(2,-2)](t[i])))**2
    # print "   : Reflection symmetry constraint (Q22,Q2-2) with interpolation", constraint1/len(t)    # error per point 
    # print "   : Example  of complex conjugate quantities in interpolation ", rholms_intp['H1'][(2,2)](0.), rholms_intp['H1'][(2,-2)](0.)


    print(" ======= rholm test: Recover the SNR of the injection at the injection parameters  ==========")
    print("  WARNING: the interpolation is done with its own epoch ")
    for det in detectors:
        lnLData = SingleDetectorLogLikelihoodData(theEpochFiducial,rholms_intp, Psig.tref, Psig.phi, Psig.theta, Psig.incl, Psig.phiref, Psig.psi, Psig.dist, 2, det)
        print(det, lnLData, np.sqrt(lnLData), rhoExpected[det])


    print(" ======= End to end LogL: Recover the SNR of the injection at the injection parameters (*)  ==========")
    print("    *** CANNOT RUN WITH FAKE DETECTORS (as in this test) **** ")
#    lnL = FactoredLogLikelihood(P, rholms_intp, crossTerms, Lmax)
#    print "  : Evan's code : ", lnL, " versus rho^2/2 ", rho2Net/2




    print(" ======= rholm test: Plot the lnLdata timeseries at the injection parameters (*)  ==========")
    # tvals = np.linspace(-0.1,0.1,5000)
    # for det in detectors:
    #     lnLData = map( (lambda x: SingleDetectorLogLikelihoodData(theEpochFiducial,rholms_intp, theEpochFiducial+x, Psig.phi,Psig.theta, Psig.incl, Psig.phiref,Psig.psi, Psig.dist, 2, det)), tvals)
    #     lnLDataEstimate = np.ones(len(tvals))*rho2Net
    #     plt.figure(1)
    #     plt.plot(tvals, lnLData,label='Ldata(t)+'+det)
    #     plt.plot(tvals, lnLDataEstimate,label="$rho^2$")
    #     plt.ylim(0, rho2Net*1.1)
    #     tEventRelative =float( Psig.tref - theEpochFiducial)
    # plt.plot([tEventRelative,0],[tEventRelative,rho2Net], color='k',linestyle='-')
    # plt.title("lnL (interpolated) vs all time")
    # plt.legend()

    tvals = np.linspace(tWindowExplore[0]+tEventFiducial,tWindowExplore[1]+tEventFiducial,int(0.25*fSample*(tWindowExplore[1]-tWindowExplore[0])))
    for det in detectors:
        lnLData = map( lambda x: SingleDetectorLogLikelihoodData(theEpochFiducial,rholms_intp, theEpochFiducial+x, Psig.phi, Psig.theta, Psig.incl, Psig.phiref,Psig.psi, Psig.dist, 2, det), tvals)
#        print lnLData
        lnLDataEstimate = np.ones(len(tvals))*rho2Net
        plt.figure(2)
        tvalsPlot = tvals  # time is done relative to a fiducial epoch so should just work
        plt.plot(tvalsPlot, lnLData,label='Ldata(t)+'+det)
        plt.plot(tvalsPlot, lnLDataEstimate,label="$rho^2$")
        plt.ylim(-rho2Net, rho2Net*1.1)
    tEventRelative =float( Psig.tref - theEpochFiducial)
    print(" Real time (relative to fiducial start time) ", tEventFiducial,  " and our triggering time is the same ", tEventRelative)
    plt.plot([tEventFiducial,tEventFiducial],[0,rho2Net], color='k',linestyle='-')
    plt.plot([tEventRelative,tEventFiducial],[0,rho2Net], color='k',linestyle='--')
    plt.title("lnLData (interpolated) vs narrow time interval")
    plt.xlabel("t(s)")
    plt.ylabel("lnL")
    plt.xlim(-0.1,0.1)
    plt.legend()
    plt.show()



    if checkResultsSlowChecks:

        print(" ======= Plotting rholm timeseries (NOT timeeshifted; are rho's offset correctly?)  ==========")
        # plot the raw rholms
        plt.figure(5)
        for det in detectors:
            QestimateScalar  = crossTerms[det][((2,2),(2,2))]*(100/distanceFiducial)/2. * np.sqrt(5./(4*lal.LAL_PI))
            rhonow = lalsim.SphHarmTimeSeriesGetMode(rholms[det], 2, 2)
            print("  rho timeseries :(det, max, expected) = ",  det,np.max(np.abs(rhonow.data.data)),QestimateScalar, "   [should be equal]")
            npts = len(rhonow.data.data)
            t = rhonow.deltaT*np.arange(npts)
            plt.plot(t,np.abs(rhonow.data.data),label='rholm(2,2):'+det)
            Qestimate = np.ones(len(t))*np.abs(QestimateScalar)
            plt.plot(t,Qestimate)
            plt.title("rholm(2,2) vs all time")
            plt.legend()
        plt.show()
        print(" ======= rholm test: interpolation check (2,2) mode: data vs timesampling ==========")
        constraint1 = 0
        for det in detectors:
            hxx = lalsim.SphHarmTimeSeriesGetMode(rholms[det], 2, 2)
            npts = len(hxx.data.data)
            t= hxx.deltaT*np.arange(npts)
        #        t = map(lambda x: x if x<npts*hxx.deltaT/2 else x-npts*hxx.deltaT, hxx.deltaT*np.arange(npts))  # center at t=0
            for i in np.arange(len(hxx.data.data)):
                constraint1+= np.abs(hxx.data.data[i]-rholms_intp[det][(2,2)](t[i]))**2
            print("   : Quality of interpolation per point : 0 ~= ", constraint1/len(hxx.data.data))






from lalsimutils import *
import emcee

rosUseGaussian = False

Tbuffer =16
fSample = 4096
deltaT = 1./fSample
psd = lalsim.SimNoisePSDiLIGOSRD

fakeSNR = 10
mtot0 = 10.
P= ChooseWaveformParams()
P.m1 = mtot0/2. * lal.LAL_MSUN_SI
P.m2 =P.m1
P.dist = 1e6*lal.LAL_PC_SI
deltaF = findDeltaF(P)

IP = Overlap(fLow=30,  deltaF=deltaF, fNyq=1./deltaT/2.)
hf1 = norm_hoff(P,IP)

P2 = P.copy()

def lnprob(x):
    global P, P2, IP, hf1
    mtot = x[0]
    if rosUseGaussian:
        ret = -0.5 * np.sum( (x-mtot0)**2/2)
    else:
        P2.m1 = mtot/2. * lal.LAL_MSUN_SI
        P2.m2 =P2.m1
        hf2 = norm_hoff(P2, IP)
        ret = np.exp(- fakeSNR*fakeSNR*(1-IP.ip(hf1,hf2)))
    print(x, ret)
    return ret

ndim, nwalkers = 1,6
p0 = [ np.array([mtot0]) + 1e-4*np.random.random(ndim) for i in range(nwalkers)]

sampler = emcee.EnsembleSampler(nwalkers, ndim, lnprob)
sampler.run_mcmc(p0,30)

for i in  sampler.flatchain:
    print(i[0])





# GOAL
#  -  visualize several samplers, to illustrate
#     how we approximate the likelihood in our sampler but preserve generality
#
# PLAN
#  - uniform sampler: make a histogram and
from lalsimutils import *
from matplotlib import pylab as plt

import RIFT.integrators.mcsampler as mcsampler
sampler = mcsampler.MCSampler()
# Sampling distribution
def uniform_samp(a, b, x):
   if type(x) is float:
       return 1/(b-a)
   else:
       return np.ones(x.shape[0])/(b-a)
import functools
# set up bounds on parameters
# Polarization angle
psi_min, psi_max = 0, 2*np.pi
# RA and dec
ra_min, ra_max = 0, 2*np.pi
dec_min, dec_max = -np.pi/2, np.pi/2
# Reference time
tref_min, tref_max = -0.1, 0.1
# Inclination angle
inc_min, inc_max = -np.pi/2, np.pi/2
# orbital phi
phi_min, phi_max = 0, 2*np.pi
# distance
dist_min, dist_max = 0.1, 100
# Uniform sampling, auto-cdf inverse
sampler.add_parameter("psi", functools.partial(uniform_samp, psi_min, psi_max), None, psi_min, psi_max)
sampler.add_parameter("ra", functools.partial(uniform_samp, ra_min, ra_max), None, ra_min, ra_max)
sampler.add_parameter("dec", functools.partial(uniform_samp, dec_min, dec_max), None, dec_min, dec_max)
sampler.add_parameter("tref", functools.partial(uniform_samp, tref_min, tref_max), None, tref_min, tref_max)
sampler.add_parameter("phi", functools.partial(uniform_samp, phi_min, phi_max), None, phi_min, phi_max)
sampler.add_parameter("inc", functools.partial(uniform_samp, inc_min, inc_max), None, inc_min, inc_max)
sampler.add_parameter("dist", functools.partial(uniform_samp, dist_min, dist_max), None, dist_min, dist_max)


ptarray = sampler.draw( 10, ("tref", "ra", "dec", "inc", "phi", "psi", "d"))
print(ptarray)
plt.hist(ptarray,50,normed=1)
plt.show()





import sys
import math
#import bisect
from collections import defaultdict
from types import ModuleType

import numpy
np=numpy #import numpy as np
from scipy import integrate, interpolate, special
import itertools
import functools



from RIFT.integrators.mcsampler_generic import MCSamplerGeneric

class ResamplingOracle(MCSamplerGeneric):
    def __init__(self,**kwargs):
        super().__init__(**kwargs)
        self.reference_samples = None
        self.reference_params = None

    def add_parameter(self, params, pdf,   **kwargs):
        super().add_parameter(params,pdf, **kwargs)

    def setup(self, reference_samples=None, reference_params=None,**kwargs):
        super().setup(**kwargs)
        self.reference_samples = reference_samples
        self.reference_params = reference_params

        # parameters that match our list.  
        # Note: we may need to initialize this LATER, if we have not yet been passed our params!
        if self.params_ordered and self.reference_params:
            print(self.params_ordered, self.reference_params)
            self.valid_params = [p for p in self.reference_params  if p in self.params_ordered] # valid parameters to sample from
            self.other_params = list( set(self.params_ordered) - set(self.valid_params)) # remainder, will be uniform

    def update_sampling_prior(self, *args, **kwargs):
        True


    def draw_simplified(self, n_samples, *args, **kwargs):
        rv_out = np.empty( (n_samples, len(self.params_ordered)))
        drawn_indx = np.random.choice(range(len(self.reference_samples)), size=n_samples) # random samples
        # Fill in raw data
        for p in self.valid_params:
            indx_out = self.params_ordered.index(p)
            indx_in = self.reference_params.index(p)
            rv_out[:,indx_out]  =  self.reference_samples[drawn_indx][:,indx_in] # copy column
        # fill in other columns
        for p in self.other_params:
            indx_out = self.params_ordered.index(p)
            rv_out[:,indx_out] = np.random.uniform(low=self.llim[p], high=self.rlim[p], size=n_samples)

        return None, None,rv_out  # compatible interface with mcsampler
        
